Package: MacroFilters
Title: Robust Trend-Cycle Decomposition for Macroeconomic Time Series
Version: 0.2.1
Authors@R: 
    person("Michal", "Kinel", , "michal.kinel@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0009-0007-3295-7199"))
Description: Provides high-performance tools for macroeconomic trend
    extraction and filtering, specifically designed to solve the end-point
    problem in real-time. Implements the MacroBoost Hybrid (MBH) filter
    using penalized P-splines and gradient boosting. Unlike the standard
    Hodrick-Prescott filter, 'MacroFilters' utilizes component-wise
    L2-boosting with robust loss functions (Huber) to handle extreme
    transient shocks (e.g., COVID-19) without inducing spurious trend
    shifts. The algorithm includes an automated two-layer diagnostic stage
    for unit roots and structural breaks, optimized via corrected AICc for
    computational efficiency. Methodology detailed in Kinel (2026)
    <doi:10.2139/ssrn.6371138>.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
LazyData: true
URL: https://github.com/michal0091/MacroFilters,
        https://michal0091.github.io/MacroFilters/
BugReports: https://github.com/michal0091/MacroFilters/issues
Imports: data.table, ggplot2, Matrix, mboost, tseries
Suggests: knitr, rmarkdown, scales, strucchange, testthat (>= 3.0.0),
        usethis, xts, zoo
VignetteBuilder: knitr
Depends: R (>= 3.5)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-11 20:35:41 UTC; miki
Author: Michal Kinel [aut, cre] (ORCID:
    <https://orcid.org/0009-0007-3295-7199>)
Maintainer: Michal Kinel <michal.kinel@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-11 21:00:02 UTC
Built: R 4.5.2; ; 2026-06-12 01:24:54 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(autoplot,macrofilter)
S3method(print,macrofilter)
export(autoplot)
export(bhp_filter)
export(hamilton_filter)
export(hp_filter)
export(mbh_filter)
importFrom(Matrix,Diagonal)
importFrom(Matrix,bandSparse)
importFrom(Matrix,crossprod)
importFrom(Matrix,solve)
importFrom(data.table,":=")
importFrom(data.table,data.table)
importFrom(data.table,setorder)
importFrom(ggplot2,aes)
importFrom(ggplot2,autoplot)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_ribbon)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_minimal)
importFrom(mboost,Huber)
importFrom(mboost,bbs)
importFrom(mboost,bols)
importFrom(mboost,boost_control)
importFrom(mboost,mboost)
importFrom(mboost,mstop)
importFrom(stats,AIC)
importFrom(stats,embed)
importFrom(stats,fitted)
importFrom(stats,frequency)
importFrom(stats,lm.fit)
importFrom(stats,mad)
importFrom(stats,qnorm)
importFrom(stats,sd)
importFrom(stats,ts)
importFrom(stats,tsp)
importFrom(tseries,adf.test)

# MacroFilters <img src="man/figures/logo.png" align="right" height="139" alt="" />

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/michal0091/MacroFilters/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/michal0091/MacroFilters/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/MacroFilters)](https://CRAN.R-project.org/package=MacroFilters)
[![DOI](https://zenodo.org/badge/DOI/10.32614/CRAN.package.MacroFilters.svg)](https://doi.org/10.32614/CRAN.package.MacroFilters)
<!-- badges: end -->

> **Working Paper Available:** The econometric methodology and
> mathematical proofs underlying this package are distributed via SSRN:
> [Robust Real-Time Macroeconomic Trend Extraction: A Gradient Boosting
> Approach](https://doi.org/10.2139/ssrn.6371138).

**MacroFilters** is a unified, high-performance library for extracting
trend and cycle components from macroeconomic time series. It combines
classical filters (Hodrick-Prescott, Hamilton, Boosted HP) with its
flagship algorithm, the **MacroBoost Hybrid (MBH)** — a
gradient-boosting filter with Huber loss that is *immune to structural
shocks* such as COVID-19, financial crises, and wars.

**Why MacroFilters instead of `mFilter` or `neverhpfilter`?**

- **Robustness:** `mbh_filter()` replaces $L_2$ squared-error loss with
  Huber loss, ensuring extreme exogenous shocks never distort the
  structural trend.
- **Speed:** The HP implementation uses sparse-matrix Cholesky
  factorisation (`Matrix`), scaling as *O(n)* instead of the dense
  *O(n³)* used by legacy packages.
- **Input agnosticism:** Pass a plain `numeric` vector, a `ts`, an
  `xts`, or a `zoo` object — the output always matches the input class
  seamlessly.

## The End-Point Problem: Solved

During extreme black swan events, traditional filters anchored in $L_2$
loss mechanically deform the long-run structural trend to absorb
massive, transitory outliers.

As demonstrated with Real US GDP during the 2020 Q2 COVID-19 collapse,
the standard HP filter bends towards the shock. The MBH filter isolates
the exogenous shock entirely within the cyclical component, preserving
absolute trend integrity in real-time.

<img src="man/figures/plot_covid_vintage.png" width="100%" />

Furthermore, ex-ante spectral alignment ensures the MBH filter perfectly
matches the baseline cyclical volatility of the industry-standard HP
filter during normal conditions, unlike the excessively volatile
Hamilton filter.

<img src="man/figures/plot_cycle_comparison.png" width="100%" />

*(Plots generated using real-time vintage data from the Federal Reserve
Economic Data - FRED).*

## Installation

``` r
# install.packages("devtools")
devtools::install_github("michal0091/MacroFilters")
```

## Quick Start Arsenal

| Function | Method | Key Advantage |
|----|----|----|
| `hp_filter()` | Hodrick-Prescott (1997) | Sparse *O(n)* implementation |
| `hamilton_filter()` | Hamilton (2018) | OLS regression, no spurious cycles |
| `bhp_filter()` | Boosted HP — Phillips & Shi (2021) | Iterative fitting with BIC/ADF stopping |
| `mbh_filter()` | MacroBoost Hybrid | Robust to outliers via Huber loss |

All functions return a list of class `c("macrofilter", "list")`.

``` r
library(MacroFilters)

# Fast, agnostic filtering on any time-series object
hp_result  <- hp_filter(us_gdp_xts)
mbh_result <- mbh_filter(us_gdp_xts)

# Access components directly
mbh_result$trend
mbh_result$cycle

# Add 95% bootstrap confidence bands and plot them
mbh_ci <- mbh_filter(us_gdp_xts, boot_iter = 50L)
autoplot(mbh_ci)            # ggplot: trend, observed series and confidence ribbon
```

## Further Reading

- `vignette("introduction", package = "MacroFilters")` — full
  walkthrough of all four filters and the S3 print/meta interface.
- `vignette("uncertainty_bands", package = "MacroFilters")` — confidence
  bands via block bootstrap and the `autoplot()` method.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)

## ----setup, message=FALSE-----------------------------------------------------
library(MacroFilters)
library(ggplot2)

## ----input-agnosticism--------------------------------------------------------
set.seed(7)
y_raw <- cumsum(rnorm(60)) + (1:60) * 0.3   # a simple integrated series

# As plain numeric
hp_num <- hp_filter(y_raw)
class(hp_num$trend)   # numeric

# As a monthly ts object
y_ts   <- ts(y_raw, start = c(2019, 1), frequency = 12)
hp_ts  <- hp_filter(y_ts)
class(hp_ts$trend)    # ts — output matches input

## ----hp-filter----------------------------------------------------------------
set.seed(42)
n  <- 100
y  <- ts(100 + 0.4 * (1:n) + 5 * sin(2 * pi * (1:n) / 20) + rnorm(n, sd = 2),
         start = c(2000, 1), frequency = 4)

hp <- hp_filter(y)
hp

## ----hp-plot, echo=FALSE, fig.height=3----------------------------------------
df_hp <- data.frame(
  t     = as.numeric(time(y)),
  data  = as.numeric(y),
  trend = as.numeric(hp$trend),
  cycle = as.numeric(hp$cycle)
)

ggplot(df_hp, aes(x = t)) +
  geom_line(aes(y = data,  colour = "Observed"), linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = trend, colour = "Trend"),    linewidth = 1.1) +
  scale_colour_manual(values = c("Observed" = "grey60", "Trend" = "#0072B2")) +
  labs(title = "HP Filter", subtitle = paste0("\u03bb = ", hp$meta$lambda),
       x = "Year", y = "Value", colour = NULL) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")

## ----hp-cycle-plot, echo=FALSE, fig.height=2.8--------------------------------
ggplot(df_hp, aes(x = t, y = cycle)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(colour = "#0072B2", linewidth = 0.8) +
  labs(title = "HP Cycle", x = "Year", y = "Cycle") +
  theme_minimal(base_size = 12)

## ----hamilton-filter----------------------------------------------------------
ham <- hamilton_filter(y)      # auto-detects h = 8 for quarterly
ham

## ----bhp-filter---------------------------------------------------------------
bhp <- bhp_filter(y, stopping = "bic")
bhp

## ----mbh-demo-----------------------------------------------------------------
set.seed(42)
n <- 80
t <- 1:n

# 1. Define simulation parameters
sd_noise <- 1.8
trend    <- 100 + 0.3 * t + 0.005 * t^2
cycle    <- 2.5 * sin(2 * pi * t / 28) # 7-year business cycle

# 2. Simulate GDP index
gdp_num <- trend + cycle + rnorm(n, sd = sd_noise)

# 3. Inject a structural shock (e.g., COVID-19 lockdown)
# Expressed as extreme standard deviation events
gdp_num[60] <- gdp_num[60] - (16 * sd_noise) # Massive crash
gdp_num[61] <- gdp_num[61] - (9 * sd_noise)  # Partial recovery
gdp_num[62] <- gdp_num[62] - (4 * sd_noise)  # V Stabilization
gdp_num[62] <- gdp_num[62] - (2 * sd_noise)  # Stabilization

gdp <- ts(gdp_num, start = c(2001, 1), frequency = 4)

# Extract trends
hp_res  <- hp_filter(gdp)

# MBH Filter: Auto-calibrated threshold (d) based on MAD of differences
mbh_res <- mbh_filter(gdp) 

mbh_res

## ----mbh-plot, echo=FALSE-----------------------------------------------------
df_mbh <- data.frame(
  t    = as.numeric(time(gdp)),
  data = as.numeric(gdp),
  hp   = as.numeric(hp_res$trend),
  mbh  = as.numeric(mbh_res$trend)
)

ggplot(df_mbh, aes(x = t)) +
  geom_line(aes(y = data, colour = "Observed"),   linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = hp,   colour = "HP trend"),   linewidth = 1.0) +
  geom_line(aes(y = mbh,  colour = "MBH trend"),  linewidth = 1.1) +
  annotate("rect",
           xmin = df_mbh$t[59], xmax = df_mbh$t[63],
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  annotate("text",
           x = df_mbh$t[61], y = max(df_mbh$data, na.rm = TRUE),
           label = "COVID shock", vjust = -0.5,
           size  = 3.5, colour = "firebrick") +
  scale_colour_manual(
    values = c("Observed" = "grey60", "HP trend" = "#0072B2", "MBH trend" = "#E69F00")
  ) +
  labs(
    title    = "HP vs MBH under a Structural Shock",
    subtitle = "MBH trend (orange) stays smooth; HP trend (blue) is pulled down",
    x = "Year", y = "GDP Index", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")

## ----s3-print-----------------------------------------------------------------
mbh_res

## ----s3-access----------------------------------------------------------------
# Trend and cycle as plain vectors
head(mbh_res$trend, 8)
head(mbh_res$cycle, 8)

# Verify the fundamental identity: trend + cycle == data
max(abs((mbh_res$trend + mbh_res$cycle) - mbh_res$data))  # should be < 1e-9

## ----s3-meta------------------------------------------------------------------
str(mbh_res$meta)

## ----cycle-comparison, fig.height=5-------------------------------------------
df_cycle <- data.frame(
  t          = as.numeric(time(gdp)),
  HP_cycle   = as.numeric(hp_res$cycle),
  MBH_cycle  = as.numeric(mbh_res$cycle)
)

ggplot(df_cycle, aes(x = t)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(aes(y = HP_cycle,  colour = "HP cycle"),  linewidth = 0.8) +
  geom_line(aes(y = MBH_cycle, colour = "MBH cycle"), linewidth = 0.8) +
  annotate("rect",
           xmin = df_cycle$t[59], xmax = df_cycle$t[63],
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  scale_colour_manual(values = c("HP cycle" = "#0072B2", "MBH cycle" = "#E69F00")) +
  labs(
    title    = "Cyclical Components",
    subtitle = "HP cycle absorbs the shock; MBH cycle faithfully records it",
    x = "Year", y = "Cycle", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")

## ----expanding-window, cache=TRUE---------------------------------------------
T_max    <- nrow(us_gdp_vintage)   # full-sample size (computed from the data)
ref_date <- as.Date("2019-10-01")  # 2019 Q4 — last pre-COVID quarter
ref_idx  <- which(us_gdp_vintage$date == ref_date)

# d calibrated on the output-gap scale (not the growth-rate scale).
# Using mad(diff(y_log)) sets d ~0.006 which is far below a typical output
# gap (0.01–0.05), so MBH over-smooths to a near-straight line.  As
# post-COVID recovery data arrive, that straight line tilts upward and
# revises the 2019 Q4 trend estimate — precisely the instability we want
# to avoid.  Fixing d = mad(hp_cycle) anchors the threshold to the right
# scale and keeps the backward revision near zero.
d_fixed <- stats::mad(hp_filter(us_gdp_vintage$gdp_log, freq = 4)$cycle)

# df = 10: avoids the extreme endpoint under-fit that df = 4 (default)
# produces.  With df = 4, the cycle at the 2019 Q4 endpoint is ~20 ppts;
# with df = 10 it is ~0.003 ppts (same as HP), removing the spurious
# upward drift that would otherwise accumulate as recovery data arrives.
# mstop = 1000 gives the additional boosting budget needed at df = 10.
fixed_df     <- 10L
fixed_mstop  <- 1000L

# Freeze the B-spline domain so every vintage uses the same basis — knot
# count and domain are set once from the full sample and never change.
fixed_knots  <- 50L                            # invariant functional space
fixed_bounds <- c(1L, T_max)                   # global B-spline domain

# Vintages: 2016 Q1 – 2022 Q4 (28 publication dates)
eval_dates   <- us_gdp_vintage[date >= "2016-01-01" & date <= "2022-10-01", date]
eval_indices <- which(us_gdp_vintage$date %in% eval_dates)

fan_list  <- vector("list", length(eval_indices))
back_list <- vector("list", length(eval_indices))

for (k in seq_along(eval_indices)) {
  i         <- eval_indices[k]
  y_current <- us_gdp_vintage$gdp_log[seq_len(i)]

  hp_res  <- hp_filter(y_current, freq = 4)

  # Single call per vintage — all parameters frozen outside the loop so
  # both the fan chart and the backward-revision series use the same basis.
  mbh_res <- mbh_filter(
    y_current,
    d              = d_fixed,
    knots          = fixed_knots,
    df             = fixed_df,
    mstop          = fixed_mstop,
    boundary.knots = fixed_bounds
  )

  n_cur    <- length(y_current)
  tail_idx <- max(1L, n_cur - 27L):n_cur   # 28-obs trailing window

  fan_list[[k]] <- data.table(
    vintage_date = us_gdp_vintage$date[i],
    obs_date     = us_gdp_vintage$date[tail_idx],
    hp_trend     = hp_res$trend[tail_idx],
    mbh_trend    = mbh_res$trend[tail_idx],
    gdp_log      = y_current[tail_idx]
  )

  if (i >= ref_idx) {
    back_list[[k]] <- data.table(
      vintage_date = us_gdp_vintage$date[i],
      hp_at_ref    = hp_res$trend[ref_idx],
      mbh_at_ref   = mbh_res$trend[ref_idx]
    )
  }
}

revisions_dt <- rbindlist(fan_list)
backward_dt  <- rbindlist(Filter(Negate(is.null), back_list))

## ----boundary-knots-demo------------------------------------------------------
n_demo  <- 200L   # truncated sample
y_demo  <- us_gdp_vintage$gdp_log[seq_len(n_demo)]

res_free      <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = NULL)
res_anchored  <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Extend by one observation and refit
y_demo_p1 <- us_gdp_vintage$gdp_log[seq_len(n_demo + 1L)]

res_free_p1     <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = NULL)
res_anchored_p1 <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Revision at the final shared observation (position n_demo)
rev_free     <- abs(res_free_p1$trend[n_demo]     - res_free$trend[n_demo])
rev_anchored <- abs(res_anchored_p1$trend[n_demo] - res_anchored$trend[n_demo])

cat(sprintf(
  "Revision at obs %d after adding one data point:\n  free domain   : %.6f\n  anchored domain: %.6f\n",
  n_demo, rev_free, rev_anchored
))

## ----fan-chart----------------------------------------------------------------
# Show only the shared overlap window (2018 Q1 onward) where every one of
# the 28 vintages contributes data — this eliminates the staircase / accordion
# artefact that arises when staggered trailing windows are plotted together.
fan_shared <- revisions_dt[obs_date >= as.Date("2018-01-01")]

p1 <- ggplot(fan_shared, aes(x = obs_date)) +
  geom_line(aes(y = hp_trend,  group = vintage_date),
            colour = "#0072B2", alpha = 0.4, linewidth = 0.6) +
  geom_line(aes(y = mbh_trend, group = vintage_date),
            colour = "#E69F00", alpha = 0.4, linewidth = 0.6) +
  geom_line(
    data = us_gdp_vintage[date >= as.Date("2018-01-01") & date <= as.Date("2022-10-01")],
    aes(x = date, y = gdp_log),
    colour = "black", linewidth = 0.8, linetype = "dashed"
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.08, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  labs(
    title    = "Vintage Fan Chart: HP (blue) vs MBH (orange)",
    subtitle = "Each line = one vintage estimate (2018–2022 overlap window). Dashed = observed.",
    x        = NULL,
    y        = "Log Real GDP"
  ) +
  theme_minimal(base_size = 12)

print(p1)

## ----backward-revision--------------------------------------------------------
# backward_dt was built in the expanding-window loop above.
# MBH used df = 10 (correct endpoint estimate), knots = 50, and
# boundary.knots = c(1, T_max) (frozen basis) — so all vintages are
# numerically comparable and the baseline trend is unbiased.
back_dt <- backward_dt[order(vintage_date)]
setnames(back_dt, c("hp_at_ref", "mbh_at_ref"), c("hp_trend", "mbh_trend"))

# Normalise to the base vintage (2019 Q4 = first vintage that includes ref_date)
base_hp  <- back_dt[vintage_date == ref_date, hp_trend]
base_mbh <- back_dt[vintage_date == ref_date, mbh_trend]

back_dt[, hp_revision  := (hp_trend  - base_hp)  * 100]   # × 100 = ppts log GDP
back_dt[, mbh_revision := (mbh_trend - base_mbh) * 100]

p2 <- ggplot(back_dt, aes(x = vintage_date)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey60") +
  geom_line(aes(y = hp_revision),  colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = hp_revision), colour = "#0072B2", size = 2.5) +
  geom_line(aes(y = mbh_revision),  colour = "#E69F00", linewidth = 1) +
  geom_point(aes(y = mbh_revision), colour = "#E69F00", size = 2.5) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$hp_revision, 1), label = "HP",
           hjust = -0.3, colour = "#0072B2", fontface = "bold") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$mbh_revision, 1), label = "MBH",
           hjust = -0.3, colour = "#E69F00", fontface = "bold") +
  scale_x_date(
    date_breaks = "6 months",
    labels      = function(x) paste0(format(x, "%Y"), " ", quarters(x)),
    expand      = expansion(add = c(30, 90))
  ) +
  labs(
    title    = "Backward Revision at 2019 Q4 as New Data Arrive",
    subtitle = "How does each filter re-estimate the pre-COVID trend as COVID data are published?",
    x        = "Vintage (publication date)",
    y        = "Revision vs 2019 Q4 baseline (log pts x100)"
  ) +
  theme_minimal(base_size = 11) +
  theme(axis.title = element_text(size = 9))

print(p2)

## ----spread-table-------------------------------------------------------------
knitr::kable(
  back_dt[, .(
    Vintage              = format(vintage_date),
    `HP trend at 2019Q4`  = round(hp_trend,  5),
    `MBH trend at 2019Q4` = round(mbh_trend, 5),
    `HP revision (ppts)`  = round(hp_revision,  3),
    `MBH revision (ppts)` = round(mbh_revision, 3)
  )],
  caption = "Backward revision of 2019 Q4 trend estimate across vintages"
)

# ── Hamilton Filter (Vectorized OLS) ──────────────────────────────────────────
# James Hamilton (2018) regression-based alternative to the HP filter.
# Regresses y_{t+h} on (1, y_t, y_{t-1}, ..., y_{t-p+1}) via OLS and defines
# the cycle as the residual.

#' Hamilton Filter
#'
#' Decomposes a time series into trend and cycle components using the
#' regression-based filter proposed by Hamilton (2018). The trend is the
#' fitted value from an OLS regression of \eqn{y_{t+h}} on
#' \eqn{(1, y_t, y_{t-1}, \ldots, y_{t-p+1})}, and the cycle is the residual.
#'
#' @param x Numeric vector, `ts`, `xts`, or `zoo` object.
#' @param h Integer horizon (number of periods ahead). If `NULL` (default),
#'   auto-detected from the series frequency using Hamilton's rule:
#'   annual = 2, quarterly = 8, monthly = 24.
#' @param p Integer number of lags in the regression (default 4).
#'
#' @details
#' Hamilton (2018) proposes replacing the HP filter with a simple regression:
#' \deqn{y_{t+h} = \beta_0 + \beta_1 y_t + \beta_2 y_{t-1} + \cdots +
#'   \beta_p y_{t-p+1} + v_{t+h}}
#' The fitted values \eqn{\hat{y}_{t+h}} define the trend and the residuals
#' \eqn{\hat{v}_{t+h}} define the cycle.
#'
#' The first \eqn{h + p - 1} observations have no computable trend or cycle
#' and are filled with `NA`.
#'
#' The lag matrix is constructed vectorized via `embed()` and the
#' regression is solved with [stats::lm.fit()] for speed.
#'
#' @return A `macrofilter` object with `trend`, `cycle`, `data`, and `meta`
#'   components. `meta` includes `h`, `p`, `coefficients`, and `compute_time`.
#'
#' @references
#' Hamilton, J.D. (2018). Why You Should Never Use the Hodrick-Prescott
#'   Filter. *Review of Economics and Statistics*, 100(5), 831--843.
#'
#' @export
#' @importFrom stats embed lm.fit
#'
#' @examples
#' # Quarterly GDP-like series
#' y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
#' result <- hamilton_filter(y)
#' print(result)
hamilton_filter <- function(x, h = NULL, p = 4L) {

  # 1. Ingest ----------------------------------------------------------------
  inputs <- ensure_computable(x)
  y <- inputs$y
  n <- length(y)

  # 2. Auto-h (horizon) ------------------------------------------------------
  if (is.null(h)) {
    if (!is.null(inputs$tsp)) {
      freq_detected <- inputs$tsp[3L]
    } else {
      freq_detected <- 4
      warning(
        "Cannot determine series frequency; assuming quarterly (freq = 4). ",
        "Pass `h` explicitly to silence this warning.",
        call. = FALSE
      )
    }
    h <- switch(as.character(freq_detected),
      "1"  = 2L,
      "4"  = 8L,
      "12" = 24L,
      as.integer(2 * freq_detected)
    )
  }
  h <- as.integer(h)
  p <- as.integer(p)

  # 3. Length validation ------------------------------------------------------
  min_obs <- h + p
  if (n <= min_obs) {
    stop(
      sprintf(
        "Series too short for selected horizon/lags (n = %d, need > %d = h + p).",
        n, min_obs
      ),
      call. = FALSE
    )
  }

  # 4. Construct regression matrix (vectorized) -------------------------------
  t0 <- proc.time()

  # embed(y, p) gives an (n - p + 1) x p matrix.
  # Row i corresponds to time t = p + i - 1 (1-based).
  # Column 1 = y_t, column 2 = y_{t-1}, ..., column p = y_{t-p+1}.
  E <- embed(y, p)

  # Keep only rows where the target y[t+h] exists: t + h <= n
  # Row i maps to t = p + i - 1, so we need i <= n - h - p + 1.
  n_valid <- n - h - p + 1L
  X <- cbind(1, E[seq_len(n_valid), , drop = FALSE])

  # Target: y[t+h] for t = p, p+1, ..., n-h
  target <- y[(p + h):n]

  # 5. OLS via lm.fit --------------------------------------------------------
  fit <- lm.fit(X, target)

  # 6. Reconstruct with NA padding --------------------------------------------
  trend_num <- rep(NA_real_, n)
  cycle_num <- rep(NA_real_, n)

  # Fitted values / residuals correspond to times (p + h) through n
  fitted_idx <- (p + h):n
  trend_num[fitted_idx] <- fit$fitted.values
  cycle_num[fitted_idx] <- fit$residuals

  elapsed <- (proc.time() - t0)[["elapsed"]]

  # 7. Package into macrofilter -----------------------------------------------
  result <- new_macrofilter(
    cycle = cycle_num,
    trend = trend_num,
    data  = y,
    meta  = list(
      method       = "Hamilton",
      h            = h,
      p            = p,
      coefficients = fit$coefficients,
      compute_time = elapsed
    )
  )
  validate_macrofilter(result)

  result
}

# ── Tests: hamilton_filter() ───────────────────────────────────────────────────

# ── 1. Identity: trend + cycle == data (non-NA positions) ─────────────────────

test_that("Hamilton trend + cycle reconstructs original data", {
  set.seed(1)
  y <- cumsum(rnorm(100))
  result <- hamilton_filter(y, h = 8, p = 4)
  valid <- !is.na(result$trend)
  expect_equal(
    result$trend[valid] + result$cycle[valid],
    result$data[valid],
    tolerance = 1e-9
  )
})

test_that("Identity holds for ts input with auto-h", {
  set.seed(2)
  y_ts <- ts(cumsum(rnorm(80)), start = c(2000, 1), frequency = 4)
  result <- hamilton_filter(y_ts)
  valid <- !is.na(result$trend)
  expect_equal(
    result$trend[valid] + result$cycle[valid],
    result$data[valid],
    tolerance = 1e-9
  )
})

# ── 2. Invariance: same result regardless of input class ─────────────────────

test_that("Hamilton filter produces identical results across input classes", {
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  y_ts  <- ts(y_num, start = c(2000, 1), frequency = 4)

  res_num <- hamilton_filter(y_num, h = 8, p = 4)
  res_ts  <- hamilton_filter(y_ts, h = 8, p = 4)

  expect_equal(res_num$trend, res_ts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_ts$cycle, tolerance = 1e-12)
})

test_that("Hamilton filter produces identical results with xts input", {
  skip_if_not_installed("xts")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_xts <- xts::xts(y_num, order.by = dates)

  res_num <- hamilton_filter(y_num, h = 8, p = 4)
  res_xts <- hamilton_filter(y_xts, h = 8, p = 4)

  expect_equal(res_num$trend, res_xts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_xts$cycle, tolerance = 1e-12)
})

test_that("Hamilton filter produces identical results with zoo input", {
  skip_if_not_installed("zoo")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_zoo <- zoo::zoo(y_num, order.by = dates)

  res_num <- hamilton_filter(y_num, h = 8, p = 4)
  res_zoo <- hamilton_filter(y_zoo, h = 8, p = 4)

  expect_equal(res_num$trend, res_zoo$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_zoo$cycle, tolerance = 1e-12)
})

# ── 3. Stress tests ──────────────────────────────────────────────────────────

test_that("Hamilton filter errors when series is too short", {
  y_short <- 1:12 * 1.0  # n = 12, need > h + p = 12
  expect_error(hamilton_filter(y_short, h = 8, p = 4), "too short")
  expect_error(hamilton_filter(1:5 * 1.0, h = 2, p = 4), "too short")
})

test_that("Hamilton filter works at minimum viable length", {
  # n = 13, h = 8, p = 4 -> n > h + p = 12 -> n_valid = n - h - p + 1 = 2
  y <- 1:13 * 1.0
  result <- hamilton_filter(y, h = 8, p = 4)
  valid <- !is.na(result$trend)
  expect_equal(sum(valid), 2L)
  expect_equal(
    result$trend[valid] + result$cycle[valid],
    result$data[valid],
    tolerance = 1e-9
  )
})

test_that("Hamilton filter handles long series (n = 10000)", {
  set.seed(99)
  y <- cumsum(rnorm(10000))
  result <- hamilton_filter(y, h = 8, p = 4)
  valid <- !is.na(result$trend)
  expect_equal(
    result$trend[valid] + result$cycle[valid],
    result$data[valid],
    tolerance = 1e-9
  )
})

test_that("First h + p - 1 positions are NA, rest are computed", {
  set.seed(10)
  y <- cumsum(rnorm(100))
  h <- 8L
  p <- 4L
  result <- hamilton_filter(y, h = h, p = p)

  n_na <- h + p - 1L
  expect_true(all(is.na(result$trend[1:n_na])))
  expect_true(all(is.na(result$cycle[1:n_na])))
  expect_true(all(!is.na(result$trend[(n_na + 1):length(y)])))
  expect_true(all(!is.na(result$cycle[(n_na + 1):length(y)])))
})

test_that("Auto-h for quarterly ts gives h = 8", {
  y_ts <- ts(1:50 * 1.0, frequency = 4)
  result <- hamilton_filter(y_ts)
  expect_equal(result$meta$h, 8L)
})

test_that("Auto-h for monthly ts gives h = 24", {
  y_ts <- ts(1:100 * 1.0, frequency = 12)
  result <- hamilton_filter(y_ts)
  expect_equal(result$meta$h, 24L)
})

test_that("Auto-h for annual ts gives h = 2", {
  y_ts <- ts(1:30 * 1.0, frequency = 1)
  result <- hamilton_filter(y_ts)
  expect_equal(result$meta$h, 2L)
})

test_that("Warning emitted when frequency cannot be determined", {
  y <- cumsum(rnorm(50))
  expect_warning(hamilton_filter(y), "Cannot determine series frequency")
})

# ── 4. Output format ─────────────────────────────────────────────────────────

test_that("Hamilton filter returns macrofilter object with correct meta", {
  set.seed(7)
  y <- cumsum(rnorm(100))
  result <- hamilton_filter(y, h = 8, p = 4)

  expect_s3_class(result, "macrofilter")
  expect_equal(result$meta$method, "Hamilton")
  expect_equal(result$meta$h, 8L)
  expect_equal(result$meta$p, 4L)
  expect_true(!is.null(result$meta$coefficients))
  expect_length(result$meta$coefficients, 5L)  # intercept + 4 lags
  expect_true(!is.null(result$meta$compute_time))
  expect_true(is.numeric(result$meta$compute_time))
})

test_that("Lengths of trend, cycle, and data all match", {
  set.seed(3)
  y <- cumsum(rnorm(100))
  result <- hamilton_filter(y, h = 8, p = 4)
  expect_equal(length(result$trend), length(result$data))
  expect_equal(length(result$cycle), length(result$data))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)

## ----setup, message=FALSE-----------------------------------------------------
library(MacroFilters)
library(ggplot2)
data("fr_gdp", package = "MacroFilters")
data("es_gdp", package = "MacroFilters")

## ----input-agnosticism--------------------------------------------------------
set.seed(7)
y_raw <- cumsum(rnorm(60)) + (1:60) * 0.3   # a simple integrated series

# As plain numeric
hp_num <- hp_filter(y_raw)
class(hp_num$trend)   # numeric

# As a monthly ts object
y_ts   <- ts(y_raw, start = c(2019, 1), frequency = 12)
hp_ts  <- hp_filter(y_ts)
class(hp_ts$trend)    # ts — output matches input

## ----hp-filter----------------------------------------------------------------
set.seed(42)
n  <- 100
y  <- ts(100 + 0.4 * (1:n) + 5 * sin(2 * pi * (1:n) / 20) + rnorm(n, sd = 2),
         start = c(2000, 1), frequency = 4)

hp <- hp_filter(y)
hp

## ----hp-plot, echo=FALSE, fig.height=3----------------------------------------
df_hp <- data.frame(
  t     = as.numeric(time(y)),
  data  = as.numeric(y),
  trend = as.numeric(hp$trend),
  cycle = as.numeric(hp$cycle)
)

ggplot(df_hp, aes(x = t)) +
  geom_line(aes(y = data,  colour = "Observed"), linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = trend, colour = "Trend"),    linewidth = 1.1) +
  scale_colour_manual(values = c("Observed" = "grey60", "Trend" = "#0072B2")) +
  labs(title = "HP Filter", subtitle = paste0("\u03bb = ", hp$meta$lambda),
       x = "Year", y = "Value", colour = NULL) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")

## ----hp-cycle-plot, echo=FALSE, fig.height=2.8--------------------------------
ggplot(df_hp, aes(x = t, y = cycle)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(colour = "#0072B2", linewidth = 0.8) +
  labs(title = "HP Cycle", x = "Year", y = "Cycle") +
  theme_minimal(base_size = 12)

## ----hamilton-filter----------------------------------------------------------
ham <- hamilton_filter(y)      # auto-detects h = 8 for quarterly
ham

## ----bhp-filter---------------------------------------------------------------
bhp <- bhp_filter(y, stopping = "bic")
bhp

## ----fred-download, eval=FALSE------------------------------------------------
# # FRED public endpoint — no API key needed.
# # See data-raw/intl_gdp.R for the full reproducible download script.
# read_fred <- function(id) {
#   url <- sprintf("https://fred.stlouisfed.org/graph/fredgraph.csv?id=%s", id)
#   dt  <- read.csv(url, col.names = c("date", "gdp_real"), na.strings = ".")
#   dt$date    <- as.Date(dt$date)
#   dt$gdp_log <- log(as.numeric(dt$gdp_real))
#   dt[!is.na(dt$gdp_real), ]
# }
# fr_raw <- read_fred("CLVMNACSCAB1GQFR")
# es_raw <- read_fred("CLVMNACSCAB1GQES")

## ----mbh-demo-----------------------------------------------------------------
# Apply HP + MBH per country.
# For log-level series, auto d (MAD of diff) is too tight — calibrate d on the
# cycle scale instead (see vignette "Hyperparameter Tuning for the MBH Filter").
make_trend_df <- function(raw, country) {
  dt  <- raw[raw$date >= as.Date("2000-01-01"), ]
  g   <- ts(dt$gdp_log, start = c(2000, 1), frequency = 4)
  hp  <- hp_filter(g)
  mbh <- mbh_filter(g, d = mad(hp$cycle))
  data.frame(country  = country,
             t        = as.numeric(time(g)),
             observed = as.numeric(g),
             hp       = as.numeric(hp$trend),
             mbh      = as.numeric(mbh$trend))
}

df_plot <- rbind(
  make_trend_df(fr_gdp, "France"),
  make_trend_df(es_gdp, "Spain")
)

# Keep Spain filter objects for the S3 class examples in Section 5
dt_es   <- es_gdp[es_gdp$date >= as.Date("2000-01-01"), ]
gdp     <- ts(dt_es$gdp_log, start = c(2000, 1), frequency = 4)
hp_res  <- hp_filter(gdp)
mbh_res <- mbh_filter(gdp, d = mad(hp_res$cycle))

mbh_res

## ----mbh-plot, echo=FALSE-----------------------------------------------------
ggplot(df_plot, aes(x = t)) +
  geom_line(aes(y = observed, colour = "Observed"),  linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = hp,       colour = "HP trend"),  linewidth = 1.0) +
  geom_line(aes(y = mbh,      colour = "MBH trend"), linewidth = 1.1) +
  annotate("rect",
           xmin = 2020.00, xmax = 2020.75,
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  annotate("text",
           x = 2020.375, y = Inf, vjust = 1.4,
           label = "COVID-19\n2020 Q2", size = 3.2, colour = "firebrick") +
  scale_colour_manual(
    values = c("Observed" = "grey60", "HP trend" = "#0072B2", "MBH trend" = "#E69F00")
  ) +
  facet_wrap(~country, scales = "free_y") +
  labs(
    title    = "HP vs MBH under a Structural Shock",
    subtitle = "MBH trend (orange) stays smooth;\nHP trend (blue) is pulled down by the COVID crash",
    x = "Year", y = "Log Real GDP", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom", strip.text = element_text(face = "bold"))

## ----s3-print-----------------------------------------------------------------
mbh_res

## ----s3-access----------------------------------------------------------------
# Trend and cycle as plain vectors
head(mbh_res$trend, 8)
head(mbh_res$cycle, 8)

# Verify the fundamental identity: trend + cycle == data
max(abs((mbh_res$trend + mbh_res$cycle) - mbh_res$data))  # should be < 1e-9

## ----s3-meta------------------------------------------------------------------
str(mbh_res$meta)

## ----cycle-comparison, fig.height=5-------------------------------------------
df_cycle <- data.frame(
  t          = as.numeric(time(gdp)),
  HP_cycle   = as.numeric(hp_res$cycle),
  MBH_cycle  = as.numeric(mbh_res$cycle)
)

ggplot(df_cycle, aes(x = t)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(aes(y = HP_cycle,  colour = "HP cycle"),  linewidth = 0.8) +
  geom_line(aes(y = MBH_cycle, colour = "MBH cycle"), linewidth = 0.8) +
  annotate("rect",
           xmin = 2020.00, xmax = 2020.75,
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  scale_colour_manual(values = c("HP cycle" = "#0072B2", "MBH cycle" = "#E69F00")) +
  labs(
    title    = "Cyclical Components",
    subtitle = "HP cycle absorbs the shock; MBH cycle faithfully records it",
    x = "Year", y = "Cycle", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")

# ── MacroBoost Hybrid Filter (MBH) ────────────────────────────────────────────
# The flagship algorithm. Combines P-Splines with Component-wise Gradient
# Boosting and Huber loss to achieve robustness against structural breaks.

#' MacroBoost Hybrid (MBH) Filter
#'
#' Decomposes a time series into trend and cycle using a robust boosting
#' algorithm. Unlike the HP filter, MBH uses the Huber loss function to
#' automatically downweight outliers (like the COVID-19 shock), preventing
#' them from distorting the trend.
#'
#' @param x Numeric vector, `ts`, `xts`, or `zoo` object.
#' @param knots Integer.
#'   Number of interior knots for the P-Spline.
#'   If `NULL` (default), it is calculated as `max(20, floor(n / 2))`.
#'   High knot density is required for the trend to be flexible enough.
#' @param mstop Integer.
#'   Maximum number of boosting iterations (default 500).
#'   If `select_mstop = TRUE` this is the upper bound; the actual stopping
#'   point is chosen by AICc.
#' @param d Numeric or `NULL`.
#'   The delta parameter for Huber loss. If `NULL` (default), it is
#'   auto-calibrated as the Median Absolute Deviation (MAD) of the first
#'   differences of the series (`mad(diff(y))`).
#'
#'   **Scale-mismatch warning:** When `x` is a log-level series,
#'   `diff(y)` returns inter-period growth rates (typical scale 0.001–0.02),
#'   whereas the output gap (the residual the filter must explain) has a much
#'   larger scale (typical scale 0.01–0.05). Using `mad(diff(y))` as `d`
#'   therefore sets the Huber threshold far too low: the filter treats normal
#'   business-cycle oscillations as outliers, truncates their gradients, and
#'   blocks learning. The trend becomes over-smooth and the cycle absorbs
#'   too much long-run variance.
#'
#'   **Recommendation:** For a quick preliminary calibration use
#'   `d = mad(hp_filter(x)$cycle)`, which sets the threshold on the residual
#'   scale. Supply an explicit numeric value to override the automatic
#'   fallback entirely.
#' @param nu Numeric.
#'   The learning rate (shrinkage) for boosting (default 0.1).
#' @param df Integer.
#'   Effective degrees of freedom per boosting step for the P-Spline base
#'   learner (default 4). This enforces the *weak-learner* constraint of
#'   Bühlmann & Hothorn (2007): each boosting step contributes only a small,
#'   smooth update so that the trend is built up gradually over many
#'   iterations rather than fitted in one pass.
#'
#'   **End-point instability warning:** Higher `df` values cause the B-spline
#'   basis matrix to shift drastically when the sample size changes by even
#'   one observation (the "rubber-band effect"). The last few data points pull
#'   the estimated trend non-smoothly, producing unreliable end-of-sample
#'   estimates. Keep `df = 4` (the default) unless you have a specific reason
#'   to deviate.
#' @param select_mstop Logical.
#'   If `TRUE`, the optimal number of boosting iterations is selected
#'   automatically via AICc (corrected AIC), following Bühlmann & Hothorn
#'   (2007). The `mstop` argument acts as the search upper bound. Default
#'   `FALSE`.
#'
#'   **AICc underfitting warning:** In the combination of Huber
#'   quasi-likelihood + P-splines, AICc penalises model complexity
#'   hyper-aggressively. In practice the algorithm stops at iteration ~5–15
#'   instead of the intended ~500. The resulting trend is nearly a straight
#'   line; all long-run variance is pushed into the cycle component, defeating
#'   the purpose of the filter. Treat `select_mstop = TRUE` as an
#'   experimental option and validate visually before relying on it.
#' @param boundary.knots A numeric vector of length 2 specifying the global
#'   domain for the B-spline basis (e.g., `c(1, T_max)`). If `NULL` (default),
#'   the range of `time_idx` is used. For real-time stability, fix this to the
#'   full-sample domain so the basis does not shift as the sample grows.
#'
#' @details
#' The model estimated is an additive model:
#' \deqn{y_t = \text{Linear}(t) + \text{Smooth}(t) + \epsilon_t}
#'
#' It is fitted using [mboost::mboost()] with:
#' \itemize{
#'   \item **Base Learners:** A linear time trend ([mboost::bols()]) to capture
#'     the global path, plus a B-spline ([mboost::bbs()]) to capture local
#'     curvature.
#'   \item **Loss Function:** Huber loss ([mboost::Huber()]) with parameter
#'     `d`.  This is the key to robustness.
#' }
#'
#' The default parameters (`knots = n/2`, `mstop = 500`) are calibrated to
#' mimic the flexibility of a standard HP filter while retaining the robustness
#' of the Huber loss.
#'
#' @section Calibration Guidance:
#' Three failure modes were discovered through empirical stress-testing.
#' The defaults guard against all three:
#'
#' \describe{
#'   \item{1. Huber delta scale mismatch (`d`)}{
#'     The automatic fallback `mad(diff(y))` operates on the scale of
#'     growth rates, not the output gap. For log-level input this sets `d`
#'     one to two orders of magnitude too small, causing ordinary
#'     business-cycle swings to be treated as outliers. If the estimated
#'     cycle looks implausibly large or the trend is nearly linear, override
#'     with `d = mad(hp_filter(x)$cycle)` as a starting point.
#'   }
#'   \item{2. AICc underfitting (`select_mstop`)}{
#'     AICc + Huber quasi-likelihood + P-splines stops boosting at
#'     iteration ~5--15. The trend degenerates to a near-straight line and
#'     the cycle absorbs all long-run variance. Leave `select_mstop = FALSE`
#'     (the default) and set `mstop` explicitly instead.
#'   }
#'   \item{3. End-point instability (`df`)}{
#'     Values above 4 shift the B-spline basis matrix non-smoothly as
#'     the sample grows, producing a "rubber-band" distortion in the final
#'     observations. Keep `df = 4` (the default) for real-time applications.
#'   }
#' }
#'
#' @return A `macrofilter` object with `trend`, `cycle`, `data`, and `meta`
#'   components.  The `meta` list contains `method = "MBH"`, `knots`, `d`,
#'   `mstop`, `nu`, `df`, `select_mstop`, and `compute_time`.
#'
#' @export
#' @importFrom mboost mboost bbs bols Huber boost_control mstop
#' @importFrom stats fitted AIC
#' @importFrom utils capture.output
#'
#' @examples
#' # Fast example with reduced series and iterations
#' set.seed(42)
#' y <- ts(cumsum(rnorm(80)), start = c(2000, 1), frequency = 4)
#' result <- mbh_filter(y, mstop = 100L)
#' print(result)
#'
#' \donttest{
#' # Full example with default parameters
#' y2 <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
#' result2 <- mbh_filter(y2)
#' print(result2)
#' }
mbh_filter <- function(x, knots = NULL, mstop = 500L, d = NULL, nu = 0.1,
                       df = 4L, select_mstop = FALSE, boundary.knots = NULL) {

  # 1. Ingest ----------------------------------------------------------------
  inputs <- ensure_computable(x)
  y <- inputs$y
  n <- length(y)

  # 2. Length validation ------------------------------------------------------
  if (n < 6L) {
    stop("Series must have at least 6 observations for MBH filter.", call. = FALSE)
  }

  # 3. Auto-calibrate d (Huber delta) ----------------------------------------
  # Computed from the MAD of first differences: a robust, scale-invariant
  # measure of cycle-to-cycle volatility.
  if (is.null(d)) {
    d <- stats::mad(diff(y), na.rm = TRUE)
    # Fallback for perfectly flat or deterministic series (MAD == 0)
    if (d < 1e-6) d <- 0.01
  }

  # 4. Knots heuristic (aggressive) ------------------------------------------
  # High flexibility (~1 knot every 2 obs) because the Huber loss will be the
  # smoothness constraint, not the spline basis.
  if (is.null(knots)) {
    knots <- max(20L, floor(n / 2L))
  }
  # Clamp knots to avoid singular spline bases with very short series
  max_knots <- max(1L, n - 4L)
  knots <- min(as.integer(knots), max_knots)
  mstop <- as.integer(mstop)
  df    <- as.integer(df)

  # 5. Timer -----------------------------------------------------------------
  t0 <- proc.time()

  # 6. Base learners ---------------------------------------------------------
  time_idx  <- seq_len(n)
  df_boost  <- data.frame(y = y, time_idx = time_idx)
  bl_linear <- mboost::bols(time_idx, intercept = TRUE)
  bl_smooth <- mboost::bbs(time_idx, knots = knots, degree = 3,
                            differences = 2, df = df,
                            boundary.knots = boundary.knots)

  # 7. Fit -------------------------------------------------------------------
  fam  <- mboost::Huber(d = d)
  ctrl <- mboost::boost_control(mstop = mstop, nu = nu)

  utils::capture.output(suppressWarnings(
    mod <- mboost::mboost(
      y ~ bl_linear + bl_smooth,
      data = df_boost, family = fam, control = ctrl
    )
  ))

  # 7.5. Optional AICc-based mstop selection ---------------------------------
  mstop_final <- mstop
  if (select_mstop) {
    aic_obj <- tryCatch(
      stats::AIC(mod, method = "corrected"),
      error = function(e) NULL
    )
    if (!is.null(aic_obj)) {
      best_stop   <- mboost::mstop(aic_obj)
      mod         <- mod[best_stop]
      mstop_final <- as.integer(best_stop)
    }
  }

  # 8. Extract ---------------------------------------------------------------
  trend_num <- as.numeric(fitted(mod))
  cycle_num <- y - trend_num

  elapsed <- (proc.time() - t0)[["elapsed"]]

  # 9. Package into macrofilter ----------------------------------------------
  result <- new_macrofilter(
    cycle = cycle_num,
    trend = trend_num,
    data  = y,
    meta  = list(
      method        = "MBH",
      knots         = knots,
      d             = d,
      mstop         = mstop_final,
      mstop_initial = mstop,
      nu            = nu,
      df            = df,
      select_mstop   = select_mstop,
      boundary.knots = boundary.knots,
      compute_time   = elapsed
    )
  )

  # 10. Validate --------------------------------------------------------------
  validate_macrofilter(result)

  result
}

# ── Tests: mbh_filter() ────────────────────────────────────────────────────────

# ── 1. Identity: trend + cycle == data ───────────────────────────────────────

test_that("Identity holds with explicit params", {
  set.seed(1)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, knots = 20L, mstop = 50L, d = 0.01)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Identity holds with auto-knots", {
  set.seed(2)
  y <- cumsum(rnorm(200))
  result <- mbh_filter(y)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

# ── 2. Invariance: same result regardless of input class ─────────────────────

test_that("MBH produces identical results across numeric and ts", {
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  y_ts  <- ts(y_num, start = c(2000, 1), frequency = 4)

  res_num <- mbh_filter(y_num, knots = 20L, mstop = 50L, d = 0.01)
  res_ts  <- mbh_filter(y_ts,  knots = 20L, mstop = 50L, d = 0.01)

  expect_equal(res_num$trend, res_ts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_ts$cycle, tolerance = 1e-12)
})

test_that("MBH produces identical results with xts input", {
  skip_if_not_installed("xts")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_xts <- xts::xts(y_num, order.by = dates)

  res_num <- mbh_filter(y_num, knots = 20L, mstop = 50L, d = 0.01)
  res_xts <- mbh_filter(y_xts, knots = 20L, mstop = 50L, d = 0.01)

  expect_equal(res_num$trend, res_xts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_xts$cycle, tolerance = 1e-12)
})

test_that("MBH produces identical results with zoo input", {
  skip_if_not_installed("zoo")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_zoo <- zoo::zoo(y_num, order.by = dates)

  res_num <- mbh_filter(y_num, knots = 20L, mstop = 50L, d = 0.01)
  res_zoo <- mbh_filter(y_zoo, knots = 20L, mstop = 50L, d = 0.01)

  expect_equal(res_num$trend, res_zoo$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_zoo$cycle, tolerance = 1e-12)
})

# ── 3. Stress tests ──────────────────────────────────────────────────────────

test_that("MBH errors when n < 6", {
  expect_error(mbh_filter(c(1, 2, 3, 4, 5)), "at least 6")
  expect_error(mbh_filter(c(1, 2, 3)),        "at least 6")
  expect_error(mbh_filter(1),                  "at least 6")
})

test_that("MBH works with minimum-length series (n = 6, 10, 20)", {
  for (n in c(6L, 10L, 20L)) {
    y <- seq_len(n) * 1.0
    result <- mbh_filter(y, mstop = 10L)
    expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
  }
})

test_that("MBH handles long series (n = 2000)", {
  set.seed(99)
  y <- cumsum(rnorm(2000))
  result <- mbh_filter(y, mstop = 50L)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Auto-knots heuristic: n=100 gives knots = max(20, 50) = 50", {
  set.seed(50)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, mstop = 10L)
  expect_equal(result$meta$knots, 50L)
})

test_that("Custom d parameter is stored correctly in meta", {
  set.seed(51)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, d = 0.5, mstop = 10L)
  expect_equal(result$meta$d, 0.5)
})

test_that("Auto-d computes correctly for normal series", {
  set.seed(10)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, mstop = 20L)
  expect_true(is.numeric(result$meta$d))
  expect_gt(result$meta$d, 0)
})

test_that("Custom d overrides Auto-d", {
  set.seed(11)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, d = 5.5, mstop = 20L)
  expect_equal(result$meta$d, 5.5)
})

test_that("Auto-d falls back to 0.01 when MAD of diffs is zero", {
  # Perfectly linear series: diff(y) is constant, so mad(diff(y)) == 0
  y <- seq(1, 20) * 1.0
  result <- mbh_filter(y, mstop = 10L)
  expect_equal(result$meta$d, 0.01)
})

# ── 4. Output format ─────────────────────────────────────────────────────────

test_that("MBH returns macrofilter with correct meta$method = 'MBH'", {
  set.seed(7)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, mstop = 20L)

  expect_s3_class(result, "macrofilter")
  expect_equal(result$meta$method, "MBH")
})

test_that("Meta contains all params: knots, d, mstop, nu, compute_time", {
  set.seed(8)
  y <- cumsum(rnorm(100))
  result <- mbh_filter(y, knots = 15L, mstop = 30L, d = 0.02, nu = 0.05)

  expect_equal(result$meta$knots, 15L)
  expect_equal(result$meta$d, 0.02)
  expect_equal(result$meta$mstop, 30L)
  expect_equal(result$meta$nu, 0.05)
  expect_true(is.numeric(result$meta$compute_time))
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")

## ----nu-equiv-----------------------------------------------------------------
y <- us_gdp_vintage$gdp_log

res_default <- mbh_filter(y, mstop = 500,  nu = 0.10)
res_equiv   <- mbh_filter(y, mstop = 1000, nu = 0.05)

max_diff <- max(abs(res_default$trend - res_equiv$trend))
cat(sprintf("Max trend difference (mstop×nu equivalence): %.2e\n", max_diff))

## ----scale-invariance---------------------------------------------------------
y_level <- us_gdp_vintage$gdp_real   # billions USD (~20 000 scale)
y_log   <- us_gdp_vintage$gdp_log    # natural log (~10 scale)

d_level <- stats::mad(diff(y_level))
d_log   <- stats::mad(diff(y_log))

cat(sprintf("d (level series) : %.4f\n", d_level))
cat(sprintf("d (log series)   : %.6f\n", d_log))
cat(sprintf("Ratio d_level / mean(level): %.6f\n", d_level / mean(y_level)))
cat(sprintf("Ratio d_log   / mean(log)  : %.6f\n", d_log   / mean(y_log)))

## ----d-sensitivity------------------------------------------------------------
y_growth <- diff(us_gdp_vintage$gdp_log)   # quarterly log-differences

res_auto   <- mbh_filter(y_growth)
res_strict <- mbh_filter(y_growth, d = 0.005)
res_lenient <- mbh_filter(y_growth, d = 0.02)

cat(sprintf("Auto d = %.6f\n", res_auto$meta$d))

## ----d-sensitivity-plot-------------------------------------------------------
dt_growth <- data.table(
  t        = us_gdp_vintage$date[-1],
  observed = y_growth,
  auto     = res_auto$trend,
  strict   = res_strict$trend,
  lenient  = res_lenient$trend
)

dt_long <- melt(dt_growth,
                id.vars      = "t",
                measure.vars = c("auto", "strict", "lenient"),
                variable.name = "delta",
                value.name    = "trend")

# Human-readable labels
auto_label <- sprintf("Auto (d=%.4f)", res_auto$meta$d)
# data.table::melt() returns variable.name as factor; fcase() returns character.
# Assigning character to a factor column via := raises a type mismatch error,
# so coerce to character first.
dt_long[, delta := as.character(delta)]
dt_long[, delta := fcase(
  delta == "auto",    auto_label,
  delta == "strict",  "Strict (d=0.005)",
  delta == "lenient", "Lenient (d=0.020)"
)]

colour_vals <- c("#0072B2", "#009E73", "#E69F00")
names(colour_vals) <- c("Strict (d=0.005)", auto_label, "Lenient (d=0.020)")

p_d <- ggplot() +
  geom_line(
    data = dt_growth,
    aes(x = t, y = observed),
    colour = "grey70", linewidth = 0.5
  ) +
  geom_line(
    data = dt_long,
    aes(x = t, y = trend, colour = delta),
    linewidth = 0.9
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID Q2\n-9% q-o-q", vjust = 1.4,
           size = 3.2, colour = "firebrick") +
  scale_colour_manual(values = colour_vals) +
  labs(
    title    = "MBH Trend Sensitivity to Huber Delta d",
    subtitle = "Data: US quarterly GDP growth rates (log-diff)",
    x        = NULL, y = "Log-difference", colour = "d setting"
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "top")

print(p_d)

## ----benchmark, cache=TRUE----------------------------------------------------
y          <- us_gdp_vintage$gdp_log
mstop_grid <- seq(100L, 1000L, by = 100L)   # 10 evenly-spaced points

bench_dt <- rbindlist(lapply(mstop_grid, function(m) {
  t0       <- proc.time()
  res      <- mbh_filter(y, mstop = m)
  elapsed  <- (proc.time() - t0)[["elapsed"]]
  cycle_sd <- sd(res$cycle)
  data.table(
    mstop       = m,
    elapsed_sec = round(elapsed, 3),
    cycle_sd    = round(cycle_sd, 6)
  )
}))

knitr::kable(
  bench_dt,
  col.names = c("mstop", "Wall time (s)", "Cycle SD"),
  caption   = "MBH computational benchmark — US log GDP (316 obs)"
)

## ----benchmark-plot-----------------------------------------------------------
# Dual-axis layout: wall time (left) + cycle_sd convergence (right)
# Use a secondary-axis trick by normalising cycle_sd to the time scale
time_range  <- range(bench_dt$elapsed_sec)
sd_range    <- range(bench_dt$cycle_sd)
# Guard against division by zero if cycle_sd converges to a flat line
if (diff(sd_range)   < 1e-10) sd_range   <- sd_range   + c(-1e-5,  1e-5)
if (diff(time_range) < 1e-10) time_range <- time_range + c(-1e-5,  1e-5)
sd_to_time  <- function(x) (x - sd_range[1]) / diff(sd_range) * diff(time_range) + time_range[1]
time_to_sd  <- function(x) (x - time_range[1]) / diff(time_range) * diff(sd_range) + sd_range[1]

p_bench <- ggplot(bench_dt, aes(x = mstop)) +
  geom_line(aes(y = elapsed_sec), colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = elapsed_sec), colour = "#CC0000", size = 3) +
  geom_line(aes(y = sd_to_time(cycle_sd)),
            colour = "#E69F00", linewidth = 0.9, linetype = "dashed") +
  geom_point(aes(y = sd_to_time(cycle_sd)),
             colour = "#E69F00", size = 2.5) +
  scale_x_continuous(breaks = mstop_grid) +
  scale_y_continuous(
    name     = "Wall time (s)  [blue / red points]",
    sec.axis = sec_axis(~ time_to_sd(.), name = "Cycle SD  [orange dashed]",
                        labels = scales::label_number(accuracy = 0.0001))
  ) +
  labs(
    title    = "Wall Time vs Boosting Iterations",
    subtitle = "US Real GDP log level (316 obs). Cycle SD plateaus well before mstop = 500.",
    x        = "mstop"
  ) +
  theme_minimal(base_size = 12) +
  theme(
    axis.title.y.left  = element_text(colour = "#0072B2"),
    axis.title.y.right = element_text(colour = "#E69F00")
  )

print(p_bench)

## ----summary-table, echo=FALSE------------------------------------------------
summary_tbl <- data.table(
  Parameter = c("`mstop`", "`nu`", "`knots`", "`d`"),
  Default   = c("500", "0.1", "`max(20, n/2)`", "auto via MAD"),
  `When to increase` = c(
    "Publication accuracy required",
    "Very long series; computational budget tight",
    "Highly nonlinear trend",
    "Series has frequent large spikes"
  ),
  `When to decrease` = c(
    "Exploratory / fast iteration",
    "Stability preferred over speed",
    "Short series or near-linear trend",
    "Series is log-level (use `mad(hp$cycle)` instead)"
  )
)
knitr::kable(summary_tbl, caption = "MBH hyperparameter quick-reference")

# ── Boosted HP Filter (Phillips & Shi, 2021) ──────────────────────────────────
# Iteratively applies the HP filter on residuals to capture stochastic trends
# that a single HP pass misses.  Three stopping rules: BIC (default), ADF,
# and fixed iteration count.

#' Boosted HP Filter
#'
#' Iteratively applies the Hodrick-Prescott filter on residuals to better
#' capture stochastic trends.  At each iteration the HP smoother is applied to
#' the current residual and the resulting trend increment is added to the
#' cumulative trend estimate.  Iteration stops according to one of three rules:
#' BIC minimisation (default), ADF stationarity test on residuals, or a fixed
#' number of iterations.
#'
#' @param x Numeric vector, `ts`, `xts`, or `zoo` object.
#' @param lambda Smoothing parameter.
#'   If `NULL` (default), it is auto-detected using the Ravn-Uhlig rule
#'   (`6.25 * freq^4`).
#' @param iter_max Integer.
#'   Maximum number of boosting iterations (default 100).
#' @param stopping Character.
#'   Stopping rule: `"bic"` (default), `"adf"`, or `"fixed"`.
#' @param sig_level Numeric.
#'   Significance level for the ADF test when `stopping = "adf"`
#'   (default 0.05).
#' @param freq Numeric frequency override (1 = annual, 4 = quarterly,
#'   12 = monthly). Used only when `lambda` is `NULL` and the frequency cannot
#'   be inferred from `x`.
#'
#' @details
#' The boosted HP filter starts from the standard HP solution and then
#' re-applies the same HP smoother to the residual (cycle) component.  The
#' trend increment from each pass is accumulated, and the procedure stops
#' when one of the following criteria is met:
#'
#' \describe{
#'   \item{`"bic"`}{Schwarz information criterion computed as
#'     \eqn{n \log(\hat\sigma^2) + \log(n)\,\mathrm{tr}(S^m)},
#'     where \eqn{S^m} is the iterated smoother.  Iteration stops when the BIC
#'     increases relative to the previous best.}
#'   \item{`"adf"`}{Augmented Dickey-Fuller test on the residual.  Iteration
#'     stops when the residual is stationary at level `sig_level`.}
#'   \item{`"fixed"`}{Runs exactly `iter_max` iterations.}
#' }
#'
#' @return A `macrofilter` object with `trend`, `cycle`, `data`, and `meta`
#'   components.  The `meta` list contains `method = "bHP"`, `lambda`,
#'   `iterations`, `stopping_rule`, and `compute_time`.
#'
#' @references
#' Phillips, P.C.B. and Shi, Z. (2021). Boosting: Why You Can Use the HP
#'   Filter. *International Economic Review*, 62(2), 521--570.
#'
#' @export
#' @importFrom Matrix bandSparse Diagonal crossprod solve
#' @importFrom tseries adf.test
#'
#' @examples
#' # Quarterly GDP-like series
#' y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
#' result <- bhp_filter(y)
#' print(result)
bhp_filter <- function(x, lambda = NULL, iter_max = 100L,
                       stopping = c("bic", "adf", "fixed"),
                       sig_level = 0.05, freq = NULL) {

  # 1. Ingest ----------------------------------------------------------------
  inputs <- ensure_computable(x)
  y <- inputs$y
  n <- length(y)

  # 2. Length validation ------------------------------------------------------
  if (n < 3L) {
    stop("Series must have at least 3 observations for bHP filter.", call. = FALSE)
  }

  # 3. Match stopping rule ---------------------------------------------------
  stopping <- match.arg(stopping)
  iter_max <- as.integer(iter_max)

  # 4. Auto-lambda (Ravn-Uhlig heuristic) ------------------------------------
  if (is.null(lambda)) {
    if (!is.null(inputs$tsp)) {
      freq_detected <- inputs$tsp[3L]
    } else if (!is.null(freq)) {
      freq_detected <- freq
    } else {
      freq_detected <- 4
      warning(
        "Cannot determine series frequency; assuming quarterly (freq = 4). ",
        "Pass `lambda` or `freq` explicitly to silence this warning.",
        call. = FALSE
      )
    }
    lambda <- 6.25 * freq_detected^4
  }

  # 5. Build sparse HP system once -------------------------------------------
  D <- Matrix::bandSparse(
    n    = n - 2L,
    m    = n,
    k    = c(0L, 1L, 2L),
    diagonals = list(
      rep(1,  n - 2L),
      rep(-2, n - 2L),
      rep(1,  n - 2L)
    )
  )
  Q <- Matrix::Diagonal(n) + lambda * Matrix::crossprod(D)

  # 6. BIC precomputation (eigenvalues of Q) ---------------------------------
  if (stopping == "bic") {
    if (n > 5000L) {
      warning(
        "BIC stopping with n > 5000 requires O(n^3) eigendecomposition. ",
        "Consider `stopping = \"adf\"` or `stopping = \"fixed\"` for large series.",
        call. = FALSE
      )
    }
    eig <- eigen(as.matrix(Q), symmetric = TRUE, only.values = TRUE)$values
    # ratios[i] = (eig[i] - 1) / eig[i]  -- these are (1 - s_i)
    ratios <- (eig - 1) / eig
  }

  # 7. Start timer -----------------------------------------------------------
  t0 <- proc.time()

  # 8. Boosting loop ---------------------------------------------------------
  # Iteration 1
  res <- .hp_solve(y, Q)
  f_hat <- res$trend
  u     <- res$cycle

  final_iter  <- 1L
  final_trend <- f_hat
  final_cycle <- u

  if (stopping == "bic") {
    best_bic   <- n * log(sum(u^2) / n) + log(n) * (n - sum(ratios^1))
    best_iter  <- 1L
    best_trend <- f_hat
    best_cycle <- u
  }

  # Iterations 2..iter_max
  if (iter_max > 1L) {
    for (k in 2L:iter_max) {
      res_k <- .hp_solve(u, Q)
      f_hat <- f_hat + res_k$trend
      u     <- res_k$cycle

      if (stopping == "fixed") {
        final_iter  <- k
        final_trend <- f_hat
        final_cycle <- u
        next
      }

      if (stopping == "adf") {
        adf_p <- tseries::adf.test(u)$p.value
        if (adf_p < sig_level) {
          final_iter  <- k
          final_trend <- f_hat
          final_cycle <- u
          break
        }
        final_iter  <- k
        final_trend <- f_hat
        final_cycle <- u
        next
      }

      if (stopping == "bic") {
        tr_Sm <- n - sum(ratios^k)
        bic_k <- n * log(sum(u^2) / n) + log(n) * tr_Sm
        if (bic_k > best_bic) {
          # BIC increased -- stop and use previous best
          final_iter  <- best_iter
          final_trend <- best_trend
          final_cycle <- best_cycle
          break
        }
        best_bic   <- bic_k
        best_iter  <- k
        best_trend <- f_hat
        best_cycle <- u
        final_iter  <- k
        final_trend <- f_hat
        final_cycle <- u
      }
    }
  }

  elapsed <- (proc.time() - t0)[["elapsed"]]

  # 9. Package into macrofilter -----------------------------------------------
  result <- new_macrofilter(
    cycle = final_cycle,
    trend = final_trend,
    data  = y,
    meta  = list(
      method        = "bHP",
      lambda        = lambda,
      iterations    = final_iter,
      stopping_rule = stopping,
      compute_time  = elapsed
    )
  )

  # 10. Validate --------------------------------------------------------------
  validate_macrofilter(result)

  result
}

# ── Internal helper ─────────────────────────────────────────────────────────
#' Apply one HP smoothing pass given a precomputed system matrix
#'
#' @param y Numeric vector to smooth.
#' @param Q Precomputed sparse system matrix (I + lambda * D'D).
#' @return A list with `trend` and `cycle` numeric vectors.
#' @noRd
#' @keywords internal
.hp_solve <- function(y, Q) {
  trend <- as.numeric(Matrix::solve(Q, y))
  list(trend = trend, cycle = y - trend)
}

# ── Tests: bhp_filter() ────────────────────────────────────────────────────────

# ── 1. Identity: trend + cycle == data ───────────────────────────────────────

test_that("Identity holds with stopping = 'fixed'", {
  set.seed(1)
  y <- cumsum(rnorm(100))
  result <- bhp_filter(y, lambda = 1600, iter_max = 10L, stopping = "fixed")
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Identity holds with stopping = 'adf'", {
  set.seed(2)
  y <- cumsum(rnorm(100))
  result <- bhp_filter(y, lambda = 1600, stopping = "adf")
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Identity holds with stopping = 'bic'", {
  set.seed(3)
  y <- cumsum(rnorm(100))
  result <- bhp_filter(y, lambda = 1600, stopping = "bic")
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

# ── 2. Invariance: same result regardless of input class ─────────────────────

test_that("bHP produces identical results across numeric and ts", {
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  y_ts  <- ts(y_num, start = c(2000, 1), frequency = 4)

  res_num <- bhp_filter(y_num, lambda = 1600, iter_max = 5L, stopping = "fixed")
  res_ts  <- bhp_filter(y_ts, lambda = 1600, iter_max = 5L, stopping = "fixed")

  expect_equal(res_num$trend, res_ts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_ts$cycle, tolerance = 1e-12)
})

test_that("bHP produces identical results with xts input", {
  skip_if_not_installed("xts")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_xts <- xts::xts(y_num, order.by = dates)

  res_num <- bhp_filter(y_num, lambda = 1600, iter_max = 5L, stopping = "fixed")
  res_xts <- bhp_filter(y_xts, lambda = 1600, iter_max = 5L, stopping = "fixed")

  expect_equal(res_num$trend, res_xts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_xts$cycle, tolerance = 1e-12)
})

test_that("bHP produces identical results with zoo input", {
  skip_if_not_installed("zoo")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_zoo <- zoo::zoo(y_num, order.by = dates)

  res_num <- bhp_filter(y_num, lambda = 1600, iter_max = 5L, stopping = "fixed")
  res_zoo <- bhp_filter(y_zoo, lambda = 1600, iter_max = 5L, stopping = "fixed")

  expect_equal(res_num$trend, res_zoo$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_zoo$cycle, tolerance = 1e-12)
})

# ── 3. Stress tests ──────────────────────────────────────────────────────────

test_that("bHP errors when n < 3", {
  expect_error(bhp_filter(c(1, 2), lambda = 1600), "at least 3")
  expect_error(bhp_filter(1, lambda = 1600), "at least 3")
})

test_that("bHP works with minimum-length series (n = 3, 4, 5)", {
  for (n in 3:5) {
    y <- seq_len(n) * 1.0
    result <- bhp_filter(y, lambda = 1600, iter_max = 3L, stopping = "fixed")
    expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
  }
})

test_that("bHP handles long series with stopping = 'fixed'", {
  set.seed(99)
  y <- cumsum(rnorm(10000))
  result <- bhp_filter(y, lambda = 1600, iter_max = 5L, stopping = "fixed")
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("iter_max = 1 with fixed stopping equals hp_filter", {
  set.seed(10)
  y <- cumsum(rnorm(100))
  res_bhp <- bhp_filter(y, lambda = 1600, iter_max = 1L, stopping = "fixed")
  res_hp  <- hp_filter(y, lambda = 1600)
  expect_equal(res_bhp$trend, res_hp$trend, tolerance = 1e-12)
  expect_equal(res_bhp$cycle, res_hp$cycle, tolerance = 1e-12)
})

test_that("ADF stops before iter_max on stationary series", {
  set.seed(11)
  # White noise is stationary; ADF should reject unit root quickly
  y <- rnorm(200)
  result <- bhp_filter(y, lambda = 1600, iter_max = 100L, stopping = "adf")
  expect_lt(result$meta$iterations, 100L)
})

test_that("BIC selects reasonable iteration count", {
  set.seed(12)
  y <- cumsum(rnorm(200))
  result <- bhp_filter(y, lambda = 1600, iter_max = 100L, stopping = "bic")
  expect_gte(result$meta$iterations, 1L)
  expect_lte(result$meta$iterations, 100L)
})

test_that("Auto-lambda from ts frequency works correctly", {
  y_ts <- ts(1:50 * 1.0, frequency = 4)
  result <- bhp_filter(y_ts, iter_max = 3L, stopping = "fixed")
  expect_equal(result$meta$lambda, 6.25 * 4^4)
})

test_that("Warning when frequency cannot be determined", {
  y <- cumsum(rnorm(50))
  expect_warning(bhp_filter(y, iter_max = 1L, stopping = "fixed"),
                 "Cannot determine series frequency")
})

# ── 4. Output format ─────────────────────────────────────────────────────────

test_that("bHP returns macrofilter with correct meta fields", {
  set.seed(7)
  y <- cumsum(rnorm(100))
  result <- bhp_filter(y, lambda = 1600, iter_max = 5L, stopping = "fixed")

  expect_s3_class(result, "macrofilter")
  expect_equal(result$meta$method, "bHP")
  expect_equal(result$meta$lambda, 1600)
  expect_equal(result$meta$stopping_rule, "fixed")
  expect_true(is.integer(result$meta$iterations))
  expect_true(is.numeric(result$meta$compute_time))
})

test_that("iterations is integer and stopping_rule is character", {
  set.seed(8)
  y <- cumsum(rnorm(100))
  result <- bhp_filter(y, lambda = 1600, stopping = "bic")

  expect_type(result$meta$iterations, "integer")
  expect_type(result$meta$stopping_rule, "character")
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")

## ----expanding-window, cache=TRUE---------------------------------------------
T_max    <- nrow(us_gdp_vintage)   # full-sample size (316 rows as of 2025)
ref_date <- as.Date("2019-10-01")  # 2019 Q4 — last pre-COVID quarter
ref_idx  <- which(us_gdp_vintage$date == ref_date)

# d calibrated on the output-gap scale (not the growth-rate scale).
# Using mad(diff(y_log)) sets d ~0.006 which is far below a typical output
# gap (0.01–0.05), so MBH over-smooths to a near-straight line.  As
# post-COVID recovery data arrive, that straight line tilts upward and
# revises the 2019 Q4 trend estimate — precisely the instability we want
# to avoid.  Fixing d = mad(hp_cycle) anchors the threshold to the right
# scale and keeps the backward revision near zero.
d_fixed <- stats::mad(hp_filter(us_gdp_vintage$gdp_log, freq = 4)$cycle)

# df = 10: avoids the extreme endpoint under-fit that df = 4 (default)
# produces.  With df = 4, the cycle at the 2019 Q4 endpoint is ~20 ppts;
# with df = 10 it is ~0.003 ppts (same as HP), removing the spurious
# upward drift that would otherwise accumulate as recovery data arrives.
# mstop = 1000 gives the additional boosting budget needed at df = 10.
fixed_df     <- 10L
fixed_mstop  <- 1000L

# Freeze the B-spline domain so every vintage uses the same basis — knot
# count and domain are set once from the full sample and never change.
fixed_knots  <- 50L                            # invariant functional space
fixed_bounds <- c(1L, T_max)                   # global B-spline domain

# Vintages: 2016 Q1 – 2022 Q4 (28 publication dates)
eval_dates   <- us_gdp_vintage[date >= "2016-01-01" & date <= "2022-10-01", date]
eval_indices <- which(us_gdp_vintage$date %in% eval_dates)

fan_list  <- vector("list", length(eval_indices))
back_list <- vector("list", length(eval_indices))

for (k in seq_along(eval_indices)) {
  i         <- eval_indices[k]
  y_current <- us_gdp_vintage$gdp_log[seq_len(i)]

  hp_res  <- hp_filter(y_current, freq = 4)

  # Single call per vintage — all parameters frozen outside the loop so
  # both the fan chart and the backward-revision series use the same basis.
  mbh_res <- mbh_filter(
    y_current,
    d              = d_fixed,
    knots          = fixed_knots,
    df             = fixed_df,
    mstop          = fixed_mstop,
    boundary.knots = fixed_bounds
  )

  n_cur    <- length(y_current)
  tail_idx <- max(1L, n_cur - 27L):n_cur   # 28-obs trailing window

  fan_list[[k]] <- data.table(
    vintage_date = us_gdp_vintage$date[i],
    obs_date     = us_gdp_vintage$date[tail_idx],
    hp_trend     = hp_res$trend[tail_idx],
    mbh_trend    = mbh_res$trend[tail_idx],
    gdp_log      = y_current[tail_idx]
  )

  if (i >= ref_idx) {
    back_list[[k]] <- data.table(
      vintage_date = us_gdp_vintage$date[i],
      hp_at_ref    = hp_res$trend[ref_idx],
      mbh_at_ref   = mbh_res$trend[ref_idx]
    )
  }
}

revisions_dt <- rbindlist(fan_list)
backward_dt  <- rbindlist(Filter(Negate(is.null), back_list))

## ----boundary-knots-demo------------------------------------------------------
n_demo  <- 200L   # truncated sample
y_demo  <- us_gdp_vintage$gdp_log[seq_len(n_demo)]

res_free      <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = NULL)
res_anchored  <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Extend by one observation and refit
y_demo_p1 <- us_gdp_vintage$gdp_log[seq_len(n_demo + 1L)]

res_free_p1     <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = NULL)
res_anchored_p1 <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Revision at the final shared observation (position n_demo)
rev_free     <- abs(res_free_p1$trend[n_demo]     - res_free$trend[n_demo])
rev_anchored <- abs(res_anchored_p1$trend[n_demo] - res_anchored$trend[n_demo])

cat(sprintf(
  "Revision at obs %d after adding one data point:\n  free domain   : %.6f\n  anchored domain: %.6f\n",
  n_demo, rev_free, rev_anchored
))

## ----fan-chart----------------------------------------------------------------
# Show only the shared overlap window (2018 Q1 onward) where every one of
# the 28 vintages contributes data — this eliminates the staircase / accordion
# artefact that arises when staggered trailing windows are plotted together.
fan_shared <- revisions_dt[obs_date >= as.Date("2018-01-01")]

p1 <- ggplot(fan_shared, aes(x = obs_date)) +
  geom_line(aes(y = hp_trend,  group = vintage_date),
            colour = "#0072B2", alpha = 0.4, linewidth = 0.6) +
  geom_line(aes(y = mbh_trend, group = vintage_date),
            colour = "#E69F00", alpha = 0.4, linewidth = 0.6) +
  geom_line(
    data = us_gdp_vintage[date >= as.Date("2018-01-01") & date <= as.Date("2022-10-01")],
    aes(x = date, y = gdp_log),
    colour = "black", linewidth = 0.8, linetype = "dashed"
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.08, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  labs(
    title    = "Vintage Fan Chart: HP (blue) vs MBH (orange)",
    subtitle = "Each line = one vintage estimate (2018–2022 overlap window). Dashed = observed.",
    x        = NULL,
    y        = "Log Real GDP"
  ) +
  theme_minimal(base_size = 12)

print(p1)

## ----backward-revision--------------------------------------------------------
# backward_dt was built in the expanding-window loop above.
# MBH used df = 10 (correct endpoint estimate), knots = 50, and
# boundary.knots = c(1, T_max) (frozen basis) — so all vintages are
# numerically comparable and the baseline trend is unbiased.
back_dt <- backward_dt[order(vintage_date)]
setnames(back_dt, c("hp_at_ref", "mbh_at_ref"), c("hp_trend", "mbh_trend"))

# Normalise to the base vintage (2019 Q4 = first vintage that includes ref_date)
base_hp  <- back_dt[vintage_date == ref_date, hp_trend]
base_mbh <- back_dt[vintage_date == ref_date, mbh_trend]

back_dt[, hp_revision  := (hp_trend  - base_hp)  * 100]   # × 100 = ppts log GDP
back_dt[, mbh_revision := (mbh_trend - base_mbh) * 100]

p2 <- ggplot(back_dt, aes(x = vintage_date)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey60") +
  geom_line(aes(y = hp_revision),  colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = hp_revision), colour = "#0072B2", size = 2.5) +
  geom_line(aes(y = mbh_revision),  colour = "#E69F00", linewidth = 1) +
  geom_point(aes(y = mbh_revision), colour = "#E69F00", size = 2.5) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$hp_revision, 1), label = "HP",
           hjust = -0.3, colour = "#0072B2", fontface = "bold") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$mbh_revision, 1), label = "MBH",
           hjust = -0.3, colour = "#E69F00", fontface = "bold") +
  scale_x_date(
    date_breaks = "6 months",
    labels      = function(x) paste0(format(x, "%Y"), " ", quarters(x)),
    expand      = expansion(add = c(30, 90))
  ) +
  labs(
    title    = "Backward Revision at 2019 Q4 as New Data Arrive",
    subtitle = "How does each filter re-estimate the pre-COVID trend as COVID data are published?",
    x        = "Vintage (publication date)",
    y        = "Revision vs 2019 Q4 baseline (log pts x100)"
  ) +
  theme_minimal(base_size = 11) +
  theme(axis.title = element_text(size = 9))

print(p2)

## ----spread-table-------------------------------------------------------------
knitr::kable(
  back_dt[, .(
    Vintage              = format(vintage_date),
    `HP trend at 2019Q4`  = round(hp_trend,  5),
    `MBH trend at 2019Q4` = round(mbh_trend, 5),
    `HP revision (ppts)`  = round(hp_revision,  3),
    `MBH revision (ppts)` = round(mbh_revision, 3)
  )],
  caption = "Backward revision of 2019 Q4 trend estimate across vintages"
)

# ── HP Filter (Sparse Matrix Optimized) ──────────────────────────────────────
# Hodrick-Prescott filter using sparse Cholesky decomposition via the Matrix
# package.  Solves (I + lambda * D'D) t = y where D is the (n-2) x n second-
# difference matrix.

#' Hodrick-Prescott Filter (Sparse Matrix Implementation)
#'
#' Decomposes a time series into trend and cycle components by solving the
#' HP penalized least-squares problem using a sparse Cholesky factorization.
#' This avoids the dense O(n^3) inversion used by other implementations and
#' scales linearly in the number of observations.
#'
#' @param x Numeric vector, `ts`, `xts`, or `zoo` object.
#' @param lambda Smoothing parameter.
#'   If `NULL` (default), it is auto-detected using the Ravn-Uhlig rule
#'   (`6.25 * freq^4`).
#' @param freq Numeric frequency override (1 = annual, 4 = quarterly,
#'   12 = monthly). Used only when `lambda` is `NULL` and the frequency cannot
#'   be inferred from `x`.
#'
#' @details
#' The HP filter minimises
#' \deqn{\sum (y_t - \tau_t)^2 + \lambda \sum (\Delta^2 \tau_t)^2}
#' which admits the closed-form solution
#' \deqn{(I + \lambda D'D)\,\tau = y}
#' where \eqn{D} is the second-difference operator.
#'
#' The implementation builds \eqn{D} as a banded sparse matrix
#' ([Matrix::bandSparse()]) and solves the symmetric positive-definite system
#' with a sparse Cholesky decomposition ([Matrix::solve()]).
#'
#' When `lambda` is not supplied the Ravn-Uhlig (2002) rule is applied:
#' `lambda = 6.25 * freq^4`, yielding 6.25 (annual), 1600 (quarterly), and
#' 129 600 (monthly).
#'
#' @return A `macrofilter` object with `trend`, `cycle`, `data`, and `meta`
#'   components.
#'
#' @references
#' Hodrick, R.J. and Prescott, E.C. (1997). Postwar U.S. Business Cycles: An
#'   Empirical Investigation. *Journal of Money, Credit and Banking*, 29(1),
#'   1--16.
#'
#' Ravn, M.O. and Uhlig, H. (2002). On Adjusting the Hodrick-Prescott Filter
#'   for the Frequency of Observations. *Review of Economics and Statistics*,
#'   84(2), 371--376.
#'
#' @export
#' @importFrom Matrix bandSparse Diagonal crossprod solve
#'
#' @examples
#' # Quarterly GDP-like series
#' y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
#' result <- hp_filter(y)
#' print(result)
hp_filter <- function(x, lambda = NULL, freq = NULL) {

  # 1. Ingest ----------------------------------------------------------------
  inputs <- ensure_computable(x)
  y <- inputs$y
  n <- length(y)

  # 2. Length validation ------------------------------------------------------
  if (n < 3L) {
    stop("Series must have at least 3 observations for HP filter.", call. = FALSE)
  }

  # 3. Auto-lambda (Ravn-Uhlig heuristic) ------------------------------------
  if (is.null(lambda)) {
    if (!is.null(inputs$tsp)) {
      freq_detected <- inputs$tsp[3L]
    } else if (!is.null(freq)) {
      freq_detected <- freq
    } else {
      freq_detected <- 4
      warning(
        "Cannot determine series frequency; assuming quarterly (freq = 4). ",
        "Pass `lambda` or `freq` explicitly to silence this warning.",
        call. = FALSE
      )
    }
    lambda <- 6.25 * freq_detected^4
  }

  # 4-6. Sparse solve (timed) ------------------------------------------------
  t0 <- proc.time()

  # 4. Second-difference matrix D: (n-2) x n, bands at k = 0, 1, 2
  D <- Matrix::bandSparse(
    n    = n - 2L,
    m    = n,
    k    = c(0L, 1L, 2L),
    diagonals = list(
      rep(1,  n - 2L),
      rep(-2, n - 2L),
      rep(1,  n - 2L)
    )
  )

  # 5. Pentadiagonal system matrix Q = I + lambda * D'D
  Q <- Matrix::Diagonal(n) + lambda * Matrix::crossprod(D)

  # 6. Solve for trend
  trend_num <- as.numeric(Matrix::solve(Q, y))
  cycle_num <- y - trend_num

  elapsed <- (proc.time() - t0)[["elapsed"]]

  # 7. Package into macrofilter -----------------------------------------------
  result <- new_macrofilter(
    cycle = cycle_num,
    trend = trend_num,
    data  = y,
    meta  = list(
      method       = "HP",
      lambda       = lambda,
      compute_time = elapsed
    )
  )
  validate_macrofilter(result)

  result
}

# ── Tests: hp_filter() ────────────────────────────────────────────────────────

# ── 1. Identity: trend + cycle == data ───────────────────────────────────────

test_that("HP trend + cycle reconstructs original data", {
  set.seed(1)
  y <- cumsum(rnorm(100))
  result <- hp_filter(y, lambda = 1600)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Identity holds for ts input with auto-lambda", {
  set.seed(2)
  y_ts <- ts(cumsum(rnorm(80)), start = c(2000, 1), frequency = 4)
  result <- hp_filter(y_ts)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

# ── 2. Invariance: same result regardless of input class ─────────────────────

test_that("HP filter produces identical results across input classes", {
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  y_ts  <- ts(y_num, start = c(2000, 1), frequency = 4)

  res_num <- hp_filter(y_num, lambda = 1600)
  res_ts  <- hp_filter(y_ts, lambda = 1600)

  expect_equal(res_num$trend, res_ts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_ts$cycle, tolerance = 1e-12)
})

test_that("HP filter produces identical results with xts input", {
  skip_if_not_installed("xts")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_xts <- xts::xts(y_num, order.by = dates)

  res_num <- hp_filter(y_num, lambda = 1600)
  res_xts <- hp_filter(y_xts, lambda = 1600)

  expect_equal(res_num$trend, res_xts$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_xts$cycle, tolerance = 1e-12)
})

test_that("HP filter produces identical results with zoo input", {
  skip_if_not_installed("zoo")
  set.seed(42)
  y_num <- cumsum(rnorm(100))
  dates <- seq(as.Date("2000-01-01"), by = "quarter", length.out = 100)
  y_zoo <- zoo::zoo(y_num, order.by = dates)

  res_num <- hp_filter(y_num, lambda = 1600)
  res_zoo <- hp_filter(y_zoo, lambda = 1600)

  expect_equal(res_num$trend, res_zoo$trend, tolerance = 1e-12)
  expect_equal(res_num$cycle, res_zoo$cycle, tolerance = 1e-12)
})

# ── 3. Stress tests ──────────────────────────────────────────────────────────

test_that("HP filter works with minimum-length series (n = 3, 4, 5)", {
  for (n in 3:5) {
    y <- seq_len(n) * 1.0
    result <- hp_filter(y, lambda = 1600)
    expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
  }
})

test_that("HP filter errors when n < 3", {
  expect_error(hp_filter(c(1, 2), lambda = 1600), "at least 3")
  expect_error(hp_filter(1, lambda = 1600), "at least 3")
})

test_that("HP filter handles long series (n = 10000)", {
  set.seed(99)
  y <- cumsum(rnorm(10000))
  result <- hp_filter(y, lambda = 1600)
  expect_equal(result$trend + result$cycle, result$data, tolerance = 1e-9)
})

test_that("Auto-lambda from ts frequency = 4 gives lambda = 1600", {
  y_ts <- ts(1:50 * 1.0, frequency = 4)
  result <- hp_filter(y_ts)
  expect_equal(result$meta$lambda, 6.25 * 4^4)
})

test_that("Auto-lambda from ts frequency = 12 gives lambda = 129600", {
  y_ts <- ts(1:50 * 1.0, frequency = 12)
  result <- hp_filter(y_ts)
  expect_equal(result$meta$lambda, 6.25 * 12^4)
})

test_that("freq parameter overrides default when lambda is NULL", {
  y <- 1:50 * 1.0
  result <- suppressWarnings(hp_filter(y, freq = 12))
  expect_equal(result$meta$lambda, 6.25 * 12^4)
})

test_that("Warning emitted when frequency cannot be determined", {
  y <- cumsum(rnorm(50))
  expect_warning(hp_filter(y), "Cannot determine series frequency")
})

# ── 4. Output format ─────────────────────────────────────────────────────────

test_that("HP filter returns macrofilter object with correct meta", {
  set.seed(7)
  y <- cumsum(rnorm(100))
  result <- hp_filter(y, lambda = 1600)

  expect_s3_class(result, "macrofilter")
  expect_equal(result$meta$method, "HP")
  expect_equal(result$meta$lambda, 1600)
  expect_true(!is.null(result$meta$compute_time))
  expect_true(is.numeric(result$meta$compute_time))
})

test_that("Explicit lambda is stored exactly in meta", {
  y <- 1:20 * 1.0
  result <- hp_filter(y, lambda = 6.25)
  expect_equal(result$meta$lambda, 6.25)
})

#' US Real GDP — FRED Vintage
#'
#' Quarterly US Real Gross Domestic Product from the Federal Reserve Bank of
#' St. Louis (FRED) public data API (series **GDPC1**), expressed in billions
#' of chained 2017 US dollars, seasonally adjusted annual rate.
#'
#' @format A `data.table` with one row per quarter and three columns:
#' \describe{
#'   \item{`date`}{`Date`. Quarter start date (e.g. `1947-01-01` = 1947 Q1).}
#'   \item{`gdp_real`}{`numeric`. Real GDP level, billions of chained 2017 USD.}
#'   \item{`gdp_log`}{`numeric`. Natural logarithm of `gdp_real`, pre-computed
#'     for convenience.}
#' }
#'
#' @details
#' The dataset covers 1947 Q1 through the latest vintage available at download
#' time (approximately 316 rows as of 2025). Rows with `NA` in `gdp_real` are
#' excluded.
#'
#' The log-level column (`gdp_log`) is particularly useful for trend-cycle
#' decomposition because log-differences approximate quarter-on-quarter
#' percentage growth rates:
#' \deqn{\Delta \log(\text{GDP}_t) \approx g_t}
#'
#' @source
#' Federal Reserve Bank of St. Louis — FRED Economic Data,
#' series GDPC1. Downloaded via the public CSV endpoint
#' `https://fred.stlouisfed.org/graph/fredgraph.csv?id=GDPC1`.
#' See `data-raw/us_gdp_vintage.R` for the reproducible download script.
#'
#' @examples
#' data("us_gdp_vintage", package = "MacroFilters")
#' head(us_gdp_vintage)
#' plot(us_gdp_vintage$date, us_gdp_vintage$gdp_log,
#'      type = "l", xlab = "Date", ylab = "Log Real GDP",
#'      main = "US Real GDP (log level)")
"us_gdp_vintage"

# ── Data Ingestion: Adapter Pattern ───────────────────────────────────────────
# ensure_computable() normalizes any time-series input to a clean numeric vector
# plus metadata. restore_class() reverses the operation.

# ── Normalizer ───────────────────────────────────────────────────────────────

#' Normalize a time-series input to a clean numeric vector
#'
#' Detects the class of `x`, extracts values and temporal metadata, and
#' validates that the series contains no `NA` or `Inf` values.
#'
#' @param x A `numeric`, `ts`, `xts`, or `zoo` object.
#' @return A list with components:
#'   \describe{
#'     \item{y}{Clean `double` vector of values.}
#'     \item{idx}{Time index vector or `NULL` (for `numeric`/`ts`).}
#'     \item{tsp}{`tsp` attribute or `NULL` (only for `ts`).}
#'     \item{class}{Character: the original class name.}
#'     \item{xts_atts}{Named list of extra `xts` attributes, or `NULL`.}
#'   }
#' @keywords internal
#' @noRd
ensure_computable <- function(x) {
  cls <- class(x)[1L]

  handler <- switch(cls,
    "numeric" = .handle_numeric(x),
    "integer" = .handle_numeric(x),
    "ts"      = .handle_ts(x),
    "xts"     = .handle_xts(x),
    "zoo"     = .handle_zoo(x),
    # data.frame / data.table / anything else -> error
    stop(
      sprintf(
        "Unsupported input class '%s'. Pass a numeric vector, ts, xts, or zoo object.\n  Hint: for data.frames use df$column.",
        cls
      ),
      call. = FALSE
    )
  )

  .validate_values(handler$y)
  handler
}

# ── Class-specific handlers (internal) ───────────────────────────────────────

.handle_numeric <- function(x) {
  list(
    y        = as.double(x),
    idx      = NULL,
    tsp      = NULL,
    class    = "numeric",
    xts_atts = NULL
  )
}

.handle_ts <- function(x) {
  list(
    y        = as.double(x),
    idx      = NULL,
    tsp      = tsp(x),
    class    = "ts",
    xts_atts = NULL
  )
}

.handle_xts <- function(x) {
  if (!requireNamespace("zoo", quietly = TRUE)) {
    stop("Package 'zoo' is required to handle xts objects.", call. = FALSE)
  }
  # Preserve xtsAttributes (user-set metadata on the xts object)
  xa <- if (requireNamespace("xts", quietly = TRUE)) xts::xtsAttributes(x) else NULL

  list(
    y        = as.double(zoo::coredata(x)),
    idx      = zoo::index(x),
    tsp      = NULL,
    class    = "xts",
    xts_atts = xa
  )
}

.handle_zoo <- function(x) {
  if (!requireNamespace("zoo", quietly = TRUE)) {
    stop("Package 'zoo' is required to handle zoo objects.", call. = FALSE)
  }
  list(
    y        = as.double(zoo::coredata(x)),
    idx      = zoo::index(x),
    tsp      = NULL,
    class    = "zoo",
    xts_atts = NULL
  )
}

# ── Value validation ─────────────────────────────────────────────────────────

.validate_values <- function(y) {
  n_nan <- sum(is.nan(y))
  n_na  <- sum(is.na(y)) - n_nan   # is.na() counts NaN too
  n_inf <- sum(is.infinite(y))

  issues <- character(0L)
  if (n_na > 0L)  issues <- c(issues, sprintf("%d NA", n_na))
  if (n_nan > 0L) issues <- c(issues, sprintf("%d NaN", n_nan))
  if (n_inf > 0L) issues <- c(issues, sprintf("%d Inf/-Inf", n_inf))

  if (length(issues) > 0L) {
    stop(
      sprintf(
        "Input contains non-finite values (%s). Clean the series before filtering.",
        paste(issues, collapse = ", ")
      ),
      call. = FALSE
    )
  }
  invisible(NULL)
}

# ── Restorer ─────────────────────────────────────────────────────────────────

#' Restore a numeric result to the original time-series class
#'
#' @param y Numeric vector (trend or cycle result from an algorithm).
#' @param input_handler The list returned by `ensure_computable()`.
#' @return An object matching the original class of the input.
#' @importFrom stats ts tsp
#' @keywords internal
#' @noRd
restore_class <- function(y, input_handler) {
  switch(input_handler$class,
    "numeric" = as.double(y),
    "ts"      = .restore_ts(y, input_handler),
    "xts"     = .restore_xts(y, input_handler),
    "zoo"     = .restore_zoo(y, input_handler),
    as.double(y)  # fallback
  )
}

.restore_ts <- function(y, h) {
  ts(y, start = h$tsp[1L], frequency = h$tsp[3L])
}

.restore_xts <- function(y, h) {
  if (!requireNamespace("xts", quietly = TRUE)) {
    stop("Package 'xts' is required to restore xts objects.", call. = FALSE)
  }
  out <- xts::xts(y, order.by = h$idx)
  if (length(h$xts_atts) > 0L) {
    xts::xtsAttributes(out) <- h$xts_atts
  }
  out
}

.restore_zoo <- function(y, h) {
  if (!requireNamespace("zoo", quietly = TRUE)) {
    stop("Package 'zoo' is required to restore zoo objects.", call. = FALSE)
  }
  zoo::zoo(y, order.by = h$idx)
}

# ── MacroFilters S3 Class System ──────────────────────────────────────────────
# Low-level constructor, validator, and print method for the `macrofilter` class.

# ── Constructor (low-level) ──────────────────────────────────────────────────

#' Create a macrofilter object (low-level)
#'
#' @param cycle Numeric vector. The cyclical component.
#' @param trend Numeric vector. The trend component.
#' @param data  Numeric vector. The original series (immutable).
#' @param meta  A named list with at least `method` (character) and optionally
#'   `lambda`, `lags`, `d`, `iter`, `compute_time`.
#' @return An object of class `macrofilter`.
#' @keywords internal
#' @noRd
new_macrofilter <- function(cycle, trend, data, meta) {
  stopifnot(is.numeric(cycle))
  stopifnot(is.numeric(trend))
  stopifnot(is.numeric(data))
  stopifnot(is.list(meta), !is.null(meta$method))

  structure(
    list(
      cycle = cycle,
      trend = trend,
      data  = data,
      meta  = meta
    ),
    class = "macrofilter"
  )
}

# ── Validator ────────────────────────────────────────────────────────────────

#' Validate a macrofilter object
#'
#' Checks that `trend + cycle` reconstructs the original data within floating
#' point tolerance.
#'
#' @param x A `macrofilter` object.
#' @param tol Numeric tolerance for the identity check (default `1e-9`).
#' @return `x`, invisibly, if valid.
#' @keywords internal
#' @noRd
validate_macrofilter <- function(x, tol = 1e-9) {
  n_cycle <- length(x$cycle)
  n_trend <- length(x$trend)
  n_data  <- length(x$data)

  if (n_cycle != n_data || n_trend != n_data) {
    stop(
      sprintf(
        "Length mismatch: data has %d obs, trend has %d, cycle has %d.",
        n_data, n_trend, n_cycle
      ),
      call. = FALSE
    )
  }

  residual <- abs(x$trend + x$cycle - x$data)
  # Ignore positions where any component is NA (e.g. Hamilton filter pads with NAs)
  check_idx <- !is.na(x$data) & !is.na(x$trend) & !is.na(x$cycle)

  if (!any(check_idx)) {
    stop("No complete cases to validate (all positions contain NA).", call. = FALSE)
  }

  max_err <- max(residual[check_idx])

  if (max_err > tol) {
    stop(
      sprintf(
        "Identity violation: max|trend + cycle - data| = %.2e (tol = %.2e).",
        max_err, tol
      ),
      call. = FALSE
    )
  }

  invisible(x)
}

# ── Print method ─────────────────────────────────────────────────────────────

#' @export
#' @importFrom stats sd
print.macrofilter <- function(x, ...) {
  m <- x$meta
  n <- length(x$data)

  cat(sprintf("-- MacroFilter [%s] --\n", m$method))
  cat(sprintf("   Observations : %d\n", n))


  # Show relevant scalar parameters per method
  params <- setdiff(names(m), c("method", "compute_time"))
  params <- Filter(function(p) length(m[[p]]) == 1L, params)
  if (length(params) > 0L) {
    vals <- vapply(params, function(p) format(m[[p]], digits = 4), character(1))
    cat(sprintf("   Parameters   : %s\n", paste(params, vals, sep = " = ", collapse = ", ")))
  }

  # Cycle summary (skip NAs)
  cy <- x$cycle[!is.na(x$cycle)]
  if (length(cy) > 0L) {
    cat(sprintf(
      "   Cycle range  : [%s, %s]  sd = %s\n",
      format(min(cy), digits = 4),
      format(max(cy), digits = 4),
      format(sd(cy),  digits = 4)
    ))
  }

  if (!is.null(m$compute_time)) {
    cat(sprintf("   Compute time : %.3f s\n", as.numeric(m$compute_time)))
  }

  invisible(x)
}

---
title: "Solving the End-Point Problem in Real-Time"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Solving the End-Point Problem in Real-Time}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")
```

## 1  The End-Point Problem

The Hodrick-Prescott filter solves the penalised least-squares problem

$$\min_{\tau} \sum_{t=1}^{T}(y_t - \tau_t)^2 + \lambda \sum_{t=3}^{T}(\Delta^2\tau_t)^2$$

which admits the closed-form solution

$$\hat{\tau} = (I + \lambda D'D)^{-1} y$$

where $D$ is the $(T-2)\times T$ second-difference operator.  The leverage
matrix $(I + \lambda D'D)^{-1}$ is symmetric and Toeplitz in the interior
of the sample, but its boundary rows are fundamentally asymmetric: the
observation at $t = T$ has **no right-hand neighbours** to anchor the
penalty, so the smoother can shift to accommodate it almost freely.

Concretely, the COVID-19 shock in 2020 Q2 represents a roughly $-10\sigma$
residual relative to any pre-pandemic trend.  Because the economy enters
and exits the shock at the end of the available sample, the HP filter
interprets the collapse as *trend information* and bends the estimated
trend backward by several quarters — even with $\lambda = 1600$.

The **MacroBoost Hybrid (MBH) filter** guards against this via Huber loss:
observations whose residuals exceed the threshold $d$ receive down-weighted
gradient contributions, so the COVID observation cannot dominate the
objective.

---

## 2  Expanding-Window Simulation

We simulate *real-time releases* by growing the sample one quarter at a
time from **2016 Q1 through 2022 Q4** (28 vintages), applying both filters
to each vintage, and recording the trailing 28-observation window.  The
longer horizon spans pre-COVID stability, the 2020 shock, and the full
recovery — giving the revision analysis enough dynamic range to separate
HP instability from MBH stability.

Three calibration choices are locked in before the loop, each addressing
a distinct failure mode:

1. **`d` on the output-gap scale** — the default `mad(diff(y))` operates
on the growth-rate scale (~0.006), far below a typical output gap
(0.01–0.05).  Setting `d = mad(hp_cycle)` anchors the Huber threshold
to the correct scale.

2. **`df = 10` for the P-spline base learner** — the default `df = 4`
imposes a very strong *weak-learner* constraint that under-fits at the
series endpoint.  For the baseline vintage (where 2019 Q4 is the last
observation), `df = 4` produces a cycle of ~20 ppts at the endpoint,
making the initial trend estimate far too low.  With `df = 10` the
endpoint cycle is ~0.003 ppts, matching HP, so subsequent vintages
do not drift upward to correct an inflated initial under-estimate.

3. **Frozen B-spline domain** — `knots = 50L` and
`boundary.knots = c(1, T_max)` are computed once so every vintage uses
an identical basis.  Without this, the default formula
`min(max(20, floor(n/2)), 250)` increases the knot count as `n` grows
(up to the cap of 250), giving later vintages progressively more spectral
flexibility.

```{r expanding-window, cache=TRUE}
T_max    <- nrow(us_gdp_vintage)   # full-sample size (computed from the data)
ref_date <- as.Date("2019-10-01")  # 2019 Q4 — last pre-COVID quarter
ref_idx  <- which(us_gdp_vintage$date == ref_date)

# d calibrated on the output-gap scale (not the growth-rate scale).
# Using mad(diff(y_log)) sets d ~0.006 which is far below a typical output
# gap (0.01–0.05), so MBH over-smooths to a near-straight line.  As
# post-COVID recovery data arrive, that straight line tilts upward and
# revises the 2019 Q4 trend estimate — precisely the instability we want
# to avoid.  Fixing d = mad(hp_cycle) anchors the threshold to the right
# scale and keeps the backward revision near zero.
d_fixed <- stats::mad(hp_filter(us_gdp_vintage$gdp_log, freq = 4)$cycle)

# df = 10: avoids the extreme endpoint under-fit that df = 4 (default)
# produces.  With df = 4, the cycle at the 2019 Q4 endpoint is ~20 ppts;
# with df = 10 it is ~0.003 ppts (same as HP), removing the spurious
# upward drift that would otherwise accumulate as recovery data arrives.
# mstop = 1000 gives the additional boosting budget needed at df = 10.
fixed_df     <- 10L
fixed_mstop  <- 1000L

# Freeze the B-spline domain so every vintage uses the same basis — knot
# count and domain are set once from the full sample and never change.
fixed_knots  <- 50L                            # invariant functional space
fixed_bounds <- c(1L, T_max)                   # global B-spline domain

# Vintages: 2016 Q1 – 2022 Q4 (28 publication dates)
eval_dates   <- us_gdp_vintage[date >= "2016-01-01" & date <= "2022-10-01", date]
eval_indices <- which(us_gdp_vintage$date %in% eval_dates)

fan_list  <- vector("list", length(eval_indices))
back_list <- vector("list", length(eval_indices))

for (k in seq_along(eval_indices)) {
  i         <- eval_indices[k]
  y_current <- us_gdp_vintage$gdp_log[seq_len(i)]

  hp_res  <- hp_filter(y_current, freq = 4)

  # Single call per vintage — all parameters frozen outside the loop so
  # both the fan chart and the backward-revision series use the same basis.
  mbh_res <- mbh_filter(
    y_current,
    d              = d_fixed,
    knots          = fixed_knots,
    df             = fixed_df,
    mstop          = fixed_mstop,
    boundary.knots = fixed_bounds
  )

  n_cur    <- length(y_current)
  tail_idx <- max(1L, n_cur - 27L):n_cur   # 28-obs trailing window

  fan_list[[k]] <- data.table(
    vintage_date = us_gdp_vintage$date[i],
    obs_date     = us_gdp_vintage$date[tail_idx],
    hp_trend     = hp_res$trend[tail_idx],
    mbh_trend    = mbh_res$trend[tail_idx],
    gdp_log      = y_current[tail_idx]
  )

  if (i >= ref_idx) {
    back_list[[k]] <- data.table(
      vintage_date = us_gdp_vintage$date[i],
      hp_at_ref    = hp_res$trend[ref_idx],
      mbh_at_ref   = mbh_res$trend[ref_idx]
    )
  }
}

revisions_dt <- rbindlist(fan_list)
backward_dt  <- rbindlist(Filter(Negate(is.null), back_list))
```

---

## 3  The Role of `boundary.knots`

The MBH filter uses a B-spline basis $B(t)$ evaluated at the integer
time index $t = 1, 2, \ldots, n$.  By default `bbs()` sets the knot
domain to `range(time_idx) = c(1, n)`.  As $n$ grows by one, the domain
expands, the basis columns shift, and the estimated trend from vintage $k$
and vintage $k+1$ are **not numerically comparable** — differences partly
reflect a changed basis rather than new data.

Passing `boundary.knots = fixed_bounds` (= `c(1, T_max)`) to every
vintage call fixes the domain to the full-sample range.  All vintages
then share an identical B-spline column space, so differences between
consecutive trend estimates are attributable purely to the additional
data point, not to basis drift.

```{r boundary-knots-demo}
n_demo  <- 200L   # truncated sample
y_demo  <- us_gdp_vintage$gdp_log[seq_len(n_demo)]

res_free      <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = NULL)
res_anchored  <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Extend by one observation and refit
y_demo_p1 <- us_gdp_vintage$gdp_log[seq_len(n_demo + 1L)]

res_free_p1     <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = NULL)
res_anchored_p1 <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Revision at the final shared observation (position n_demo)
rev_free     <- abs(res_free_p1$trend[n_demo]     - res_free$trend[n_demo])
rev_anchored <- abs(res_anchored_p1$trend[n_demo] - res_anchored$trend[n_demo])

cat(sprintf(
  "Revision at obs %d after adding one data point:\n  free domain   : %.6f\n  anchored domain: %.6f\n",
  n_demo, rev_free, rev_anchored
))
```

The anchored revision is smaller because a fixed `boundary.knots` removes
a numerical artefact that would otherwise masquerade as genuine trend
movement.

---

## 4  Vintage Fan Chart

Each semi-transparent line below is one vintage estimate.  A **wide fan**
(many non-overlapping lines) means the filter is revising its trend view
strongly as new data arrive — a sign of end-point instability.  A **tight
bundle** means the filter is stable in real time.

```{r fan-chart}
# Show only the shared overlap window (2018 Q1 onward) where every one of
# the 28 vintages contributes data — this eliminates the staircase / accordion
# artefact that arises when staggered trailing windows are plotted together.
fan_shared <- revisions_dt[obs_date >= as.Date("2018-01-01")]

p1 <- ggplot(fan_shared, aes(x = obs_date)) +
  geom_line(aes(y = hp_trend,  group = vintage_date),
            colour = "#0072B2", alpha = 0.4, linewidth = 0.6) +
  geom_line(aes(y = mbh_trend, group = vintage_date),
            colour = "#E69F00", alpha = 0.4, linewidth = 0.6) +
  geom_line(
    data = us_gdp_vintage[date >= as.Date("2018-01-01") & date <= as.Date("2022-10-01")],
    aes(x = date, y = gdp_log),
    colour = "black", linewidth = 0.8, linetype = "dashed"
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.08, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  labs(
    title    = "Vintage Fan Chart: HP (blue) vs MBH (orange)",
    subtitle = "Each line = one vintage estimate (2018–2022 overlap window). Dashed = observed.",
    x        = NULL,
    y        = "Log Real GDP"
  ) +
  theme_minimal(base_size = 12)

print(p1)
```

The blue HP fan opens visibly around the 2020 COVID quarters — the filter
revises its trend estimate substantially as new post-COVID data become
available.  The orange MBH bundle remains tight: the Huber loss prevents
the COVID shock from distorting the trend, so successive vintages produce
nearly identical estimates.

---

## 5  The Backward Revision Test

The most direct proof of end-point instability is not the *current*
trend estimate but **how a past estimate gets rewritten** as new data
arrive.  Pick a fixed pre-COVID reference date — 2019 Q4 — and ask:
*how does each filter's trend estimate at that date change as we add
2020–2022 data?*

For HP, the COVID shock at the end of the sample bends the smoother
backward: 2019 Q4's trend is revised *downward* when COVID arrives
(HP interprets the collapse as trend), then partially corrected as
recovery data is added.  For MBH, the Huber loss caps the influence of
the outlier, so earlier trend estimates barely move.

```{r backward-revision}
# backward_dt was built in the expanding-window loop above.
# MBH used df = 10 (correct endpoint estimate), knots = 50, and
# boundary.knots = c(1, T_max) (frozen basis) — so all vintages are
# numerically comparable and the baseline trend is unbiased.
back_dt <- backward_dt[order(vintage_date)]
setnames(back_dt, c("hp_at_ref", "mbh_at_ref"), c("hp_trend", "mbh_trend"))

# Normalise to the base vintage (2019 Q4 = first vintage that includes ref_date)
base_hp  <- back_dt[vintage_date == ref_date, hp_trend]
base_mbh <- back_dt[vintage_date == ref_date, mbh_trend]

back_dt[, hp_revision  := (hp_trend  - base_hp)  * 100]   # × 100 = ppts log GDP
back_dt[, mbh_revision := (mbh_trend - base_mbh) * 100]

p2 <- ggplot(back_dt, aes(x = vintage_date)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey60") +
  geom_line(aes(y = hp_revision),  colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = hp_revision), colour = "#0072B2", size = 2.5) +
  geom_line(aes(y = mbh_revision),  colour = "#E69F00", linewidth = 1) +
  geom_point(aes(y = mbh_revision), colour = "#E69F00", size = 2.5) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$hp_revision, 1), label = "HP",
           hjust = -0.3, colour = "#0072B2", fontface = "bold") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$mbh_revision, 1), label = "MBH",
           hjust = -0.3, colour = "#E69F00", fontface = "bold") +
  scale_x_date(
    date_breaks = "6 months",
    labels      = function(x) paste0(format(x, "%Y"), " ", quarters(x)),
    expand      = expansion(add = c(30, 90))
  ) +
  labs(
    title    = "Backward Revision at 2019 Q4 as New Data Arrive",
    subtitle = "How does each filter re-estimate the pre-COVID trend as COVID data are published?",
    x        = "Vintage (publication date)",
    y        = "Revision vs 2019 Q4 baseline (log pts x100)"
  ) +
  theme_minimal(base_size = 11) +
  theme(axis.title = element_text(size = 9))

print(p2)
```

HP (blue) dips sharply when 2020 Q2 data arrive — the filter
re-attributes pre-COVID growth as "trend decline" — then partially
recovers as post-COVID data accumulate.  MBH (orange) dips roughly half
as much: the Huber loss limits how much the COVID observation can drag the
estimated pre-COVID trend downward, so the backward revision is
substantially smaller than HP's throughout the simulation window.

```{r spread-table}
knitr::kable(
  back_dt[, .(
    Vintage              = format(vintage_date),
    `HP trend at 2019Q4`  = round(hp_trend,  5),
    `MBH trend at 2019Q4` = round(mbh_trend, 5),
    `HP revision (ppts)`  = round(hp_revision,  3),
    `MBH revision (ppts)` = round(mbh_revision, 3)
  )],
  caption = "Backward revision of 2019 Q4 trend estimate across vintages"
)
```

---

## Summary

| Property | HP Filter | MBH Filter |
|---|---|---|
| End-point leverage | High (boundary asymmetry) | Low (Huber down-weighting) |
| Backward revision of pre-COVID trend | Large (dips on shock) | Near-zero |
| COVID trend distortion | Severe | Minimal |
| Basis stability across vintages | N/A | Requires `boundary.knots` |

For real-time monitoring applications — nowcasting output gaps, setting
automatic stabilisers, informing monetary policy decisions — the MBH
filter's robustness substantially reduces the risk that a preliminary
trend estimate is later revised beyond recognition.

---
title: "Introduction to MacroFilters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to MacroFilters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)
```

```{r setup, message=FALSE}
library(MacroFilters)
library(ggplot2)
data("fr_gdp", package = "MacroFilters")
data("es_gdp", package = "MacroFilters")
```

---

## 1. Introduction: Trend-Cycle Decomposition

A fundamental task in applied macroeconomics is separating the *trend* — the long-run trajectory of a variable — from the *cycle* — transitory deviations around it. This decomposition underpins business-cycle analysis, the output gap, and potential GDP estimation.

Every series can be written as:

$$
y_t = \tau_t + c_t
$$

where $\tau_t$ is the trend and $c_t$ is the cyclical component. The challenge is that any filter must decide whether an unusual observation represents a *genuine shock to the trend* or a *transitory deviation that belongs in the cycle*.

### The outlier problem

Classical filters minimise squared loss. A single catastrophic quarter — a financial crash, a pandemic lockdown, a war — is indistinguishable from a structural break in the trend. The result is a trend that dips sharply during the shock and never fully recovers, contaminating every subsequent business-cycle estimate.

**MacroFilters** solves this with the `mbh_filter()` function, which uses Huber loss to automatically down-weight extreme observations while fitting a smooth trend via gradient boosting.

---

## 2. Input Agnosticism: Bring Your Own Class

Many filter packages force you to convert data to a specific time-series class before calling them. **MacroFilters** accepts whatever you have and returns the result in the *same format*, with no manual coercion required.

Supported input classes:

| Class | Package | Example |
|---|---|---|
| `numeric` | base R | `c(100, 102, 98, ...)` |
| `ts` | base R | `ts(y, start = c(2000, 1), frequency = 4)` |
| `xts` | xts | `xts(y, order.by = dates)` |
| `zoo` | zoo | `zoo(y, order.by = dates)` |

### Example: same filter, two input formats

```{r input-agnosticism}
set.seed(7)
y_raw <- cumsum(rnorm(60)) + (1:60) * 0.3   # a simple integrated series

# As plain numeric
hp_num <- hp_filter(y_raw)
class(hp_num$trend)   # numeric

# As a monthly ts object
y_ts   <- ts(y_raw, start = c(2019, 1), frequency = 12)
hp_ts  <- hp_filter(y_ts)
class(hp_ts$trend)    # ts — output matches input
```

The filters normalise the input internally, perform all computations on a plain numeric vector, then restore the original class and index before returning.

---

## 3. The Filter Arsenal

### 3.1 `hp_filter()` — Sparse Hodrick-Prescott

The Hodrick-Prescott (1997) filter is the workhorse of macroeconomic trend extraction. It solves the penalised least-squares problem:

$$
\min_{\tau} \sum_{t=1}^{n}(y_t - \tau_t)^2 + \lambda \sum_{t=2}^{n-1}[(\tau_{t+1} - \tau_t) - (\tau_t - \tau_{t-1})]^2
$$

The second term penalises curvature in the trend; $\lambda$ controls the smoothness.

Most implementations solve this by inverting a dense $n \times n$ matrix, which is $O(n^3)$. **MacroFilters** recognises that the penalty matrix $D'D$ is *pentadiagonal* (a sparse banded structure) and solves the system using `Matrix::bandSparse()` and sparse Cholesky factorisation — bringing the cost down to **O(n)** in time and memory.

```{r hp-filter}
set.seed(42)
n  <- 100
y  <- ts(100 + 0.4 * (1:n) + 5 * sin(2 * pi * (1:n) / 20) + rnorm(n, sd = 2),
         start = c(2000, 1), frequency = 4)

hp <- hp_filter(y)
hp
```

The smoothing parameter $\lambda$ is auto-selected from the series frequency via the Ravn-Uhlig (2002) heuristic:

$$
\lambda = 6.25 \times \text{freq}^4
$$

which gives $\lambda = 1{,}600$ for quarterly and $\lambda = 129{,}600$ for monthly data — the conventional values. You can override it explicitly: `hp_filter(y, lambda = 1600)`.

```{r hp-plot, echo=FALSE, fig.height=3}
df_hp <- data.frame(
  t     = as.numeric(time(y)),
  data  = as.numeric(y),
  trend = as.numeric(hp$trend),
  cycle = as.numeric(hp$cycle)
)

ggplot(df_hp, aes(x = t)) +
  geom_line(aes(y = data,  colour = "Observed"), linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = trend, colour = "Trend"),    linewidth = 1.1) +
  scale_colour_manual(values = c("Observed" = "grey60", "Trend" = "#0072B2")) +
  labs(title = "HP Filter", subtitle = paste0("\u03bb = ", hp$meta$lambda),
       x = "Year", y = "Value", colour = NULL) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")
```

```{r hp-cycle-plot, echo=FALSE, fig.height=2.8}
ggplot(df_hp, aes(x = t, y = cycle)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(colour = "#0072B2", linewidth = 0.8) +
  labs(title = "HP Cycle", x = "Year", y = "Cycle") +
  theme_minimal(base_size = 12)
```

---

### 3.2 `hamilton_filter()` — Regression-Based Alternative

Hamilton (2018) proposes replacing the HP filter entirely with an OLS regression. The idea: project $y_{t+h}$ on a constant and $p$ lags of $y_t$:

$$
y_{t+h} = \alpha_0 + \alpha_1 y_t + \alpha_2 y_{t-1} + \cdots + \alpha_p y_{t-p+1} + c_t
$$

The fitted values form the trend; the residuals form the cycle. The horizon $h$ is set to two years ahead by default (e.g., $h = 8$ quarters), long enough to capture business-cycle variation without filtering it out.

**Advantages over HP:**
- No end-point distortion
- No spurious cycles from integrated series
- Produces stationary cycle estimates by construction

```{r hamilton-filter}
ham <- hamilton_filter(y)      # auto-detects h = 8 for quarterly
ham
```

Note that the first $h + p - 1$ observations of the trend and cycle are `NA`, because there is insufficient history for the regression.

---

### 3.3 `bhp_filter()` — Boosted HP

Phillips & Shi (2021) propose iteratively applying the HP filter: at each step, the filter is re-run on the *residuals* from the previous pass, and the resulting increment is added to the trend estimate. This procedure converges to a trend that better tracks stochastic variation.

$$
\tau^{(m)} = \tau^{(m-1)} + S_\lambda \cdot c^{(m-1)}, \qquad c^{(m)} = y - \tau^{(m)}
$$

where $S_\lambda$ is the HP smoothing operator. Three stopping rules are available:

| Rule | Description |
|---|---|
| `"bic"` (default) | Minimise the Schwarz information criterion |
| `"adf"` | Stop when the cycle passes an Augmented Dickey-Fuller stationarity test |
| `"fixed"` | Run exactly `iter_max` iterations |

```{r bhp-filter}
bhp <- bhp_filter(y, stopping = "bic")
bhp
```

Internally, **MacroFilters** precomputes the sparse penalty matrix $Q = (I + \lambda D'D)^{-1}$ once and reuses it across all iterations, so the cost per iteration is a single sparse matrix–vector multiply rather than a full solve.

---

## 4. The Crown Jewel: `mbh_filter()`

### The Problem with Squared Loss

Every filter above minimises squared residuals. When a pandemic shock drops GDP by 15% in a single quarter, that one observation exerts enormous leverage — it is 100× more influential than a typical quarterly fluctuation. The filter accommodates it by *bending the trend downward*, producing a spurious trend break that infects every subsequent output gap estimate.

### The MBH Solution: Huber Loss + Boosting

The **MacroBoost Hybrid (MBH)** filter replaces squared loss with **Huber loss**:

$$
L_\delta(r) = \begin{cases}
\dfrac{r^2}{2} & |r| \leq \delta \\[6pt]
\delta\!\left(|r| - \dfrac{\delta}{2}\right) & |r| > \delta
\end{cases}
$$

Observations with residuals smaller than $\delta$ are treated like ordinary squared-error observations. Observations with residuals larger than $\delta$ — the structural shocks — contribute only *linearly*, massively reducing their influence on the trend estimate.

### Additive Model

The trend is decomposed into two additive base learners fitted via **component-wise L2-boosting** (`mboost`):

$$
\hat{\tau}_t = \underbrace{b_0 + b_1 \cdot t}_{\text{Global linear drift}} + \underbrace{f(t)}_{\text{Local curvature (P-spline)}}
$$

- **`bols(time, intercept = TRUE)`** — captures the global linear drift.
- **`bbs(time, knots, degree = 3, differences = 2, boundary.knots)`** — a cubic P-spline with `knots` interior knots that captures smooth nonlinear curvature.

The default `knots = min(max(20, floor(n/2)), 250)` is deliberately generous — it gives the spline enough flexibility to follow genuine low-frequency movements without overfitting, while the Huber loss ensures that shock-contaminated quarters are downweighted. The cap of 250 keeps the basis bounded on long / high-frequency series, where extra knots add cost but not flexibility (the penalty, not the knot count, governs smoothness).

### Parameters

| Parameter | Default | Role |
|---|---|---|
| `knots` | `min(max(20, n/2), 250)` | P-spline flexibility — higher = more local adaptability (capped at 250) |
| `mstop` | `500` | Boosting iterations — more = finer approximation |
| `d` | `"auto"` | Huber delta — if `"auto"`, calibrated via the MAD of the HP cyclical residual (output-gap scale) |
| `nu` | `0.1` | Shrinkage / learning rate — controls step size per iteration |
| `boundary.knots` | `NULL` | B-spline domain anchor — if `NULL`, uses `range(time_idx)`; fix to `c(1, T_max)` for real-time vintage stability |

By default, `d` is auto-calibrated as the MAD of the HP cyclical residual, anchoring the Huber threshold to the output-gap (business-cycle) scale rather than the growth-rate scale. You can override it with an explicit numeric value: `d = 0.01` is typical for growth rates, while larger values suit index-level series.

For real-time applications where the sample grows one period at a time,
set `boundary.knots = c(1, T_max)` to anchor the B-spline domain to the
full-sample range — this prevents the basis from shifting between vintages
and makes trends comparable across publication dates.

### Quick example

France and Spain had two of the sharpest COVID-19 contractions in the EU (approximately −14 % and −18 % quarter-on-quarter in 2020 Q2 respectively), both followed by a rapid V-shaped recovery — making them a demanding real-world stress test for any trend filter.

```{r fred-download, eval=FALSE}
# FRED public endpoint — no API key needed.
# See data-raw/intl_gdp.R for the full reproducible download script.
read_fred <- function(id) {
  url <- sprintf("https://fred.stlouisfed.org/graph/fredgraph.csv?id=%s", id)
  dt  <- read.csv(url, col.names = c("date", "gdp_real"), na.strings = ".")
  dt$date    <- as.Date(dt$date)
  dt$gdp_log <- log(as.numeric(dt$gdp_real))
  dt[!is.na(dt$gdp_real), ]
}
fr_raw <- read_fred("CLVMNACSCAB1GQFR")
es_raw <- read_fred("CLVMNACSCAB1GQES")
```

```{r mbh-demo}
# Apply HP + MBH per country.
# For log-level series, auto d (MAD of diff) is too tight — calibrate d on the
# cycle scale instead (see vignette "Hyperparameter Tuning for the MBH Filter").
make_trend_df <- function(raw, country) {
  dt  <- raw[raw$date >= as.Date("2000-01-01"), ]
  g   <- ts(dt$gdp_log, start = c(2000, 1), frequency = 4)
  hp  <- hp_filter(g)
  mbh <- mbh_filter(g, d = mad(hp$cycle))
  data.frame(country  = country,
             t        = as.numeric(time(g)),
             observed = as.numeric(g),
             hp       = as.numeric(hp$trend),
             mbh      = as.numeric(mbh$trend))
}

df_plot <- rbind(
  make_trend_df(fr_gdp, "France"),
  make_trend_df(es_gdp, "Spain")
)

# Keep Spain filter objects for the S3 class examples in Section 5
dt_es   <- es_gdp[es_gdp$date >= as.Date("2000-01-01"), ]
gdp     <- ts(dt_es$gdp_log, start = c(2000, 1), frequency = 4)
hp_res  <- hp_filter(gdp)
mbh_res <- mbh_filter(gdp, d = mad(hp_res$cycle))

mbh_res
```

```{r mbh-plot, echo=FALSE}
ggplot(df_plot, aes(x = t)) +
  geom_line(aes(y = observed, colour = "Observed"),  linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = hp,       colour = "HP trend"),  linewidth = 1.0) +
  geom_line(aes(y = mbh,      colour = "MBH trend"), linewidth = 1.1) +
  annotate("rect",
           xmin = 2020.00, xmax = 2020.75,
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  annotate("text",
           x = 2020.375, y = Inf, vjust = 1.4,
           label = "COVID-19\n2020 Q2", size = 3.2, colour = "firebrick") +
  scale_colour_manual(
    values = c("Observed" = "grey60", "HP trend" = "#0072B2", "MBH trend" = "#E69F00")
  ) +
  facet_wrap(~country, scales = "free_y") +
  labs(
    title    = "HP vs MBH under a Structural Shock",
    subtitle = "MBH trend (orange) stays smooth;\nHP trend (blue) is pulled down by the COVID crash",
    x = "Year", y = "Log Real GDP", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom", strip.text = element_text(face = "bold"))
```

---

## 5. The `macrofilter` S3 Class

All four functions return a list of class `c("macrofilter", "list")` with four
core named elements:

| Element | Type | Description |
|---|---|---|
| `$trend` | numeric | Trend component |
| `$cycle` | numeric | Cyclical component |
| `$data` | numeric | Original series (immutable) |
| `$meta` | named list | Filter method, parameters, temporal identity (`ts_class`, `tsp`, `idx`), compute time |

When a filter is called with `boot_iter > 0`, the object additionally carries
two bootstrap confidence bands (see the *Uncertainty bands* vignette):

| Element | Type | Description |
|---|---|---|
| `$trend_lower` | numeric | Lower 95% band (`trend - 1.96 * bootstrap sd`) |
| `$trend_upper` | numeric | Upper 95% band (`trend + 1.96 * bootstrap sd`) |

### Printing

```{r s3-print}
mbh_res
```

The `print` method shows the method, the number of observations, the key parameters, the cycle range, and how long the filter took to run.

### Accessing components

```{r s3-access}
# Trend and cycle as plain vectors
head(mbh_res$trend, 8)
head(mbh_res$cycle, 8)

# Verify the fundamental identity: trend + cycle == data
max(abs((mbh_res$trend + mbh_res$cycle) - mbh_res$data))  # should be < 1e-9
```

### Inspecting metadata

```{r s3-meta}
str(mbh_res$meta)
```

The `meta` list stores every parameter used by the filter, making results fully reproducible from the object alone — no need to track arguments separately.

### Plotting cycles side by side

```{r cycle-comparison, fig.height=5}
df_cycle <- data.frame(
  t          = as.numeric(time(gdp)),
  HP_cycle   = as.numeric(hp_res$cycle),
  MBH_cycle  = as.numeric(mbh_res$cycle)
)

ggplot(df_cycle, aes(x = t)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(aes(y = HP_cycle,  colour = "HP cycle"),  linewidth = 0.8) +
  geom_line(aes(y = MBH_cycle, colour = "MBH cycle"), linewidth = 0.8) +
  annotate("rect",
           xmin = 2020.00, xmax = 2020.75,
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  scale_colour_manual(values = c("HP cycle" = "#0072B2", "MBH cycle" = "#E69F00")) +
  labs(
    title    = "Cyclical Components",
    subtitle = "HP cycle absorbs the shock; MBH cycle faithfully records it",
    x = "Year", y = "Cycle", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")
```

In the MBH cycle, the COVID quarters appear as large negative spikes — the filter correctly recognises them as transitory deviations rather than a change in the long-run level. The HP cycle, by contrast, spreads the shock over several surrounding quarters as it tries to reconcile the trend distortion.

---

## References

- Hodrick, R. J., & Prescott, E. C. (1997). Postwar U.S. Business Cycles: An Empirical Investigation. *Journal of Money, Credit and Banking*, 29(1), 1–16.
- Ravn, M. O., & Uhlig, H. (2002). On Adjusting the Hodrick-Prescott Filter for the Frequency of Observations. *Review of Economics and Statistics*, 84(2), 371–376.
- Hamilton, J. D. (2018). Why You Should Never Use the Hodrick-Prescott Filter. *Review of Economics and Statistics*, 100(5), 831–843.
- Phillips, P. C. B., & Shi, Z. (2021). Boosting: Why You Can Use the HP Filter. *International Economic Review*, 62(2), 521–570.

---
title: "Introduction to MacroFilters"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to MacroFilters}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)
```

```{r setup, message=FALSE}
library(MacroFilters)
library(ggplot2)
```

---

## 1. Introduction: Trend-Cycle Decomposition

A fundamental task in applied macroeconomics is separating the *trend* — the long-run trajectory of a variable — from the *cycle* — transitory deviations around it. This decomposition underpins business-cycle analysis, the output gap, and potential GDP estimation.

Every series can be written as:

$$
y_t = \tau_t + c_t
$$

where $\tau_t$ is the trend and $c_t$ is the cyclical component. The challenge is that any filter must decide whether an unusual observation represents a *genuine shock to the trend* or a *transitory deviation that belongs in the cycle*.

### The outlier problem

Classical filters minimise squared loss. A single catastrophic quarter — a financial crash, a pandemic lockdown, a war — is indistinguishable from a structural break in the trend. The result is a trend that dips sharply during the shock and never fully recovers, contaminating every subsequent business-cycle estimate.

**MacroFilters** solves this with the `mbh_filter()` function, which uses Huber loss to automatically down-weight extreme observations while fitting a smooth trend via gradient boosting.

---

## 2. Input Agnosticism: Bring Your Own Class

Many filter packages force you to convert data to a specific time-series class before calling them. **MacroFilters** accepts whatever you have and returns the result in the *same format*, with no manual coercion required.

Supported input classes:

| Class | Package | Example |
|---|---|---|
| `numeric` | base R | `c(100, 102, 98, ...)` |
| `ts` | base R | `ts(y, start = c(2000, 1), frequency = 4)` |
| `xts` | xts | `xts(y, order.by = dates)` |
| `zoo` | zoo | `zoo(y, order.by = dates)` |

### Example: same filter, two input formats

```{r input-agnosticism}
set.seed(7)
y_raw <- cumsum(rnorm(60)) + (1:60) * 0.3   # a simple integrated series

# As plain numeric
hp_num <- hp_filter(y_raw)
class(hp_num$trend)   # numeric

# As a monthly ts object
y_ts   <- ts(y_raw, start = c(2019, 1), frequency = 12)
hp_ts  <- hp_filter(y_ts)
class(hp_ts$trend)    # ts — output matches input
```

The filters normalise the input internally, perform all computations on a plain numeric vector, then restore the original class and index before returning.

---

## 3. The Filter Arsenal

### 3.1 `hp_filter()` — Sparse Hodrick-Prescott

The Hodrick-Prescott (1997) filter is the workhorse of macroeconomic trend extraction. It solves the penalised least-squares problem:

$$
\min_{\tau} \sum_{t=1}^{n}(y_t - \tau_t)^2 + \lambda \sum_{t=2}^{n-1}[(\tau_{t+1} - \tau_t) - (\tau_t - \tau_{t-1})]^2
$$

The second term penalises curvature in the trend; $\lambda$ controls the smoothness.

Most implementations solve this by inverting a dense $n \times n$ matrix, which is $O(n^3)$. **MacroFilters** recognises that the penalty matrix $D'D$ is *pentadiagonal* (a sparse banded structure) and solves the system using `Matrix::bandSparse()` and sparse Cholesky factorisation — bringing the cost down to **O(n)** in time and memory.

```{r hp-filter}
set.seed(42)
n  <- 100
y  <- ts(100 + 0.4 * (1:n) + 5 * sin(2 * pi * (1:n) / 20) + rnorm(n, sd = 2),
         start = c(2000, 1), frequency = 4)

hp <- hp_filter(y)
hp
```

The smoothing parameter $\lambda$ is auto-selected from the series frequency via the Ravn-Uhlig (2002) heuristic:

$$
\lambda = 6.25 \times \text{freq}^4
$$

which gives $\lambda = 1{,}600$ for quarterly and $\lambda = 129{,}600$ for monthly data — the conventional values. You can override it explicitly: `hp_filter(y, lambda = 1600)`.

```{r hp-plot, echo=FALSE, fig.height=3}
df_hp <- data.frame(
  t     = as.numeric(time(y)),
  data  = as.numeric(y),
  trend = as.numeric(hp$trend),
  cycle = as.numeric(hp$cycle)
)

ggplot(df_hp, aes(x = t)) +
  geom_line(aes(y = data,  colour = "Observed"), linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = trend, colour = "Trend"),    linewidth = 1.1) +
  scale_colour_manual(values = c("Observed" = "grey60", "Trend" = "#0072B2")) +
  labs(title = "HP Filter", subtitle = paste0("\u03bb = ", hp$meta$lambda),
       x = "Year", y = "Value", colour = NULL) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")
```

```{r hp-cycle-plot, echo=FALSE, fig.height=2.8}
ggplot(df_hp, aes(x = t, y = cycle)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(colour = "#0072B2", linewidth = 0.8) +
  labs(title = "HP Cycle", x = "Year", y = "Cycle") +
  theme_minimal(base_size = 12)
```

---

### 3.2 `hamilton_filter()` — Regression-Based Alternative

Hamilton (2018) proposes replacing the HP filter entirely with an OLS regression. The idea: project $y_{t+h}$ on a constant and $p$ lags of $y_t$:

$$
y_{t+h} = \alpha_0 + \alpha_1 y_t + \alpha_2 y_{t-1} + \cdots + \alpha_p y_{t-p+1} + c_t
$$

The fitted values form the trend; the residuals form the cycle. The horizon $h$ is set to two years ahead by default (e.g., $h = 8$ quarters), long enough to capture business-cycle variation without filtering it out.

**Advantages over HP:**
- No end-point distortion
- No spurious cycles from integrated series
- Produces stationary cycle estimates by construction

```{r hamilton-filter}
ham <- hamilton_filter(y)      # auto-detects h = 8 for quarterly
ham
```

Note that the first $h + p - 1$ observations of the trend and cycle are `NA`, because there is insufficient history for the regression.

---

### 3.3 `bhp_filter()` — Boosted HP

Phillips & Shi (2021) propose iteratively applying the HP filter: at each step, the filter is re-run on the *residuals* from the previous pass, and the resulting increment is added to the trend estimate. This procedure converges to a trend that better tracks stochastic variation.

$$
\tau^{(m)} = \tau^{(m-1)} + S_\lambda \cdot c^{(m-1)}, \qquad c^{(m)} = y - \tau^{(m)}
$$

where $S_\lambda$ is the HP smoothing operator. Three stopping rules are available:

| Rule | Description |
|---|---|
| `"bic"` (default) | Minimise the Schwarz information criterion |
| `"adf"` | Stop when the cycle passes an Augmented Dickey-Fuller stationarity test |
| `"fixed"` | Run exactly `iter_max` iterations |

```{r bhp-filter}
bhp <- bhp_filter(y, stopping = "bic")
bhp
```

Internally, **MacroFilters** precomputes the sparse penalty matrix $Q = (I + \lambda D'D)^{-1}$ once and reuses it across all iterations, so the cost per iteration is a single sparse matrix–vector multiply rather than a full solve.

---

## 4. The Crown Jewel: `mbh_filter()`

### The Problem with Squared Loss

Every filter above minimises squared residuals. When a pandemic shock drops GDP by 15% in a single quarter, that one observation exerts enormous leverage — it is 100× more influential than a typical quarterly fluctuation. The filter accommodates it by *bending the trend downward*, producing a spurious trend break that infects every subsequent output gap estimate.

### The MBH Solution: Huber Loss + Boosting

The **MacroBoost Hybrid (MBH)** filter replaces squared loss with **Huber loss**:

$$
L_\delta(r) = \begin{cases}
\dfrac{r^2}{2} & |r| \leq \delta \\[6pt]
\delta\!\left(|r| - \dfrac{\delta}{2}\right) & |r| > \delta
\end{cases}
$$

Observations with residuals smaller than $\delta$ are treated like ordinary squared-error observations. Observations with residuals larger than $\delta$ — the structural shocks — contribute only *linearly*, massively reducing their influence on the trend estimate.

### Additive Model

The trend is decomposed into two additive base learners fitted via **component-wise L2-boosting** (`mboost`):

$$
\hat{\tau}_t = \underbrace{b_0 + b_1 \cdot t}_{\text{Global linear drift}} + \underbrace{f(t)}_{\text{Local curvature (P-spline)}}
$$

- **`bols(time, intercept = TRUE)`** — captures the global linear drift.
- **`bbs(time, knots, degree = 3, differences = 2, boundary.knots)`** — a cubic P-spline with `knots` interior knots that captures smooth nonlinear curvature.

The default `knots = max(20, floor(n/2))` is deliberately generous — it gives the spline enough flexibility to follow genuine low-frequency movements without overfitting, while the Huber loss ensures that shock-contaminated quarters are downweighted.

### Parameters

| Parameter | Default | Role |
|---|---|---|
| `knots` | `max(20, n/2)` | P-spline flexibility — higher = more local adaptability |
| `mstop` | `500` | Boosting iterations — more = finer approximation |
| `d` | `NULL (Auto)` | Huber delta — if `NULL`, auto-set via MAD of first differences (scale-invariant) |
| `nu` | `0.2` | Shrinkage / learning rate — controls step size per iteration |
| `boundary.knots` | `NULL` | B-spline domain anchor — if `NULL`, uses `range(time_idx)`; fix to `c(1, T_max)` for real-time vintage stability |

By default, `d` is auto-calibrated using the MAD of the series' first differences, making the filter scale-invariant and suitable for both log-differenced and level series. You can override it with an explicit numeric value: `d = 0.01` is typical for growth rates, while `d = 5` to `d = 20` may be needed for index-level series.

For real-time applications where the sample grows one period at a time,
set `boundary.knots = c(1, T_max)` to anchor the B-spline domain to the
full-sample range — this prevents the basis from shifting between vintages
and makes trends comparable across publication dates.

### Quick example

```{r mbh-demo}
set.seed(42)
n <- 80
t <- 1:n

# 1. Define simulation parameters
sd_noise <- 1.8
trend    <- 100 + 0.3 * t + 0.005 * t^2
cycle    <- 2.5 * sin(2 * pi * t / 28) # 7-year business cycle

# 2. Simulate GDP index
gdp_num <- trend + cycle + rnorm(n, sd = sd_noise)

# 3. Inject a structural shock (e.g., COVID-19 lockdown)
# Expressed as extreme standard deviation events
gdp_num[60] <- gdp_num[60] - (16 * sd_noise) # Massive crash
gdp_num[61] <- gdp_num[61] - (9 * sd_noise)  # Partial recovery
gdp_num[62] <- gdp_num[62] - (4 * sd_noise)  # V Stabilization
gdp_num[62] <- gdp_num[62] - (2 * sd_noise)  # Stabilization

gdp <- ts(gdp_num, start = c(2001, 1), frequency = 4)

# Extract trends
hp_res  <- hp_filter(gdp)

# MBH Filter: Auto-calibrated threshold (d) based on MAD of differences
mbh_res <- mbh_filter(gdp) 

mbh_res
```

```{r mbh-plot, echo=FALSE}
df_mbh <- data.frame(
  t    = as.numeric(time(gdp)),
  data = as.numeric(gdp),
  hp   = as.numeric(hp_res$trend),
  mbh  = as.numeric(mbh_res$trend)
)

ggplot(df_mbh, aes(x = t)) +
  geom_line(aes(y = data, colour = "Observed"),   linewidth = 0.6, linetype = "dashed") +
  geom_line(aes(y = hp,   colour = "HP trend"),   linewidth = 1.0) +
  geom_line(aes(y = mbh,  colour = "MBH trend"),  linewidth = 1.1) +
  annotate("rect",
           xmin = df_mbh$t[59], xmax = df_mbh$t[63],
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  annotate("text",
           x = df_mbh$t[61], y = max(df_mbh$data, na.rm = TRUE),
           label = "COVID shock", vjust = -0.5,
           size  = 3.5, colour = "firebrick") +
  scale_colour_manual(
    values = c("Observed" = "grey60", "HP trend" = "#0072B2", "MBH trend" = "#E69F00")
  ) +
  labs(
    title    = "HP vs MBH under a Structural Shock",
    subtitle = "MBH trend (orange) stays smooth; HP trend (blue) is pulled down",
    x = "Year", y = "GDP Index", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")
```

---

## 5. The `macrofilter` S3 Class

All four functions return an object of class `macrofilter`. This is a list with three named elements:

| Element | Type | Description |
|---|---|---|
| `$trend` | numeric / ts / xts / zoo | Trend component |
| `$cycle` | numeric / ts / xts / zoo | Cyclical component |
| `$data` | numeric / ts / xts / zoo | Original series (immutable) |
| `$meta` | named list | Filter method, parameters, compute time |

### Printing

```{r s3-print}
mbh_res
```

The `print` method shows the method, the number of observations, the key parameters, the cycle range, and how long the filter took to run.

### Accessing components

```{r s3-access}
# Trend and cycle as plain vectors
head(mbh_res$trend, 8)
head(mbh_res$cycle, 8)

# Verify the fundamental identity: trend + cycle == data
max(abs((mbh_res$trend + mbh_res$cycle) - mbh_res$data))  # should be < 1e-9
```

### Inspecting metadata

```{r s3-meta}
str(mbh_res$meta)
```

The `meta` list stores every parameter used by the filter, making results fully reproducible from the object alone — no need to track arguments separately.

### Plotting cycles side by side

```{r cycle-comparison, fig.height=5}
df_cycle <- data.frame(
  t          = as.numeric(time(gdp)),
  HP_cycle   = as.numeric(hp_res$cycle),
  MBH_cycle  = as.numeric(mbh_res$cycle)
)

ggplot(df_cycle, aes(x = t)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey40") +
  geom_line(aes(y = HP_cycle,  colour = "HP cycle"),  linewidth = 0.8) +
  geom_line(aes(y = MBH_cycle, colour = "MBH cycle"), linewidth = 0.8) +
  annotate("rect",
           xmin = df_cycle$t[59], xmax = df_cycle$t[63],
           ymin = -Inf, ymax = Inf,
           alpha = 0.12, fill = "firebrick") +
  scale_colour_manual(values = c("HP cycle" = "#0072B2", "MBH cycle" = "#E69F00")) +
  labs(
    title    = "Cyclical Components",
    subtitle = "HP cycle absorbs the shock; MBH cycle faithfully records it",
    x = "Year", y = "Cycle", colour = NULL
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "bottom")
```

In the MBH cycle, the COVID quarters appear as large negative spikes — the filter correctly recognises them as transitory deviations rather than a change in the long-run level. The HP cycle, by contrast, spreads the shock over several surrounding quarters as it tries to reconcile the trend distortion.

---

## References

- Hodrick, R. J., & Prescott, E. C. (1997). Postwar U.S. Business Cycles: An Empirical Investigation. *Journal of Money, Credit and Banking*, 29(1), 1–16.
- Ravn, M. O., & Uhlig, H. (2002). On Adjusting the Hodrick-Prescott Filter for the Frequency of Observations. *Review of Economics and Statistics*, 84(2), 371–376.
- Hamilton, J. D. (2018). Why You Should Never Use the Hodrick-Prescott Filter. *Review of Economics and Statistics*, 100(5), 831–843.
- Phillips, P. C. B., & Shi, Z. (2021). Boosting: Why You Can Use the HP Filter. *International Economic Review*, 62(2), 521–570.

---
title: "Solving the End-Point Problem in Real-Time"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Solving the End-Point Problem in Real-Time}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")
```

## 1  The End-Point Problem

The Hodrick-Prescott filter solves the penalised least-squares problem

$$\min_{\tau} \sum_{t=1}^{T}(y_t - \tau_t)^2 + \lambda \sum_{t=3}^{T}(\Delta^2\tau_t)^2$$

which admits the closed-form solution

$$\hat{\tau} = (I + \lambda D'D)^{-1} y$$

where $D$ is the $(T-2)\times T$ second-difference operator.  The leverage
matrix $(I + \lambda D'D)^{-1}$ is symmetric and Toeplitz in the interior
of the sample, but its boundary rows are fundamentally asymmetric: the
observation at $t = T$ has **no right-hand neighbours** to anchor the
penalty, so the smoother can shift to accommodate it almost freely.

Concretely, the COVID-19 shock in 2020 Q2 represents a roughly $-10\sigma$
residual relative to any pre-pandemic trend.  Because the economy enters
and exits the shock at the end of the available sample, the HP filter
interprets the collapse as *trend information* and bends the estimated
trend backward by several quarters — even with $\lambda = 1600$.

The **MacroBoost Hybrid (MBH) filter** guards against this via Huber loss:
observations whose residuals exceed the threshold $d$ receive down-weighted
gradient contributions, so the COVID observation cannot dominate the
objective.

---

## 2  Expanding-Window Simulation

We simulate *real-time releases* by growing the sample one quarter at a
time from **2016 Q1 through 2022 Q4** (28 vintages), applying both filters
to each vintage, and recording the trailing 28-observation window.  The
longer horizon spans pre-COVID stability, the 2020 shock, and the full
recovery — giving the revision analysis enough dynamic range to separate
HP instability from MBH stability.

Three calibration choices are locked in before the loop, each addressing
a distinct failure mode:

1. **`d` on the output-gap scale** — the default `mad(diff(y))` operates
on the growth-rate scale (~0.006), far below a typical output gap
(0.01–0.05).  Setting `d = mad(hp_cycle)` anchors the Huber threshold
to the correct scale.

2. **`df = 10` for the P-spline base learner** — the default `df = 4`
imposes a very strong *weak-learner* constraint that under-fits at the
series endpoint.  For the baseline vintage (where 2019 Q4 is the last
observation), `df = 4` produces a cycle of ~20 ppts at the endpoint,
making the initial trend estimate far too low.  With `df = 10` the
endpoint cycle is ~0.003 ppts, matching HP, so subsequent vintages
do not drift upward to correct an inflated initial under-estimate.

3. **Frozen B-spline domain** — `knots = 50L` and
`boundary.knots = c(1, T_max)` are computed once so every vintage uses
an identical basis.  Without this, the default formula
`max(20, floor(n/2))` increases the knot count as `n` grows, giving
later vintages progressively more spectral flexibility.

```{r expanding-window, cache=TRUE}
T_max    <- nrow(us_gdp_vintage)   # full-sample size (316 rows as of 2025)
ref_date <- as.Date("2019-10-01")  # 2019 Q4 — last pre-COVID quarter
ref_idx  <- which(us_gdp_vintage$date == ref_date)

# d calibrated on the output-gap scale (not the growth-rate scale).
# Using mad(diff(y_log)) sets d ~0.006 which is far below a typical output
# gap (0.01–0.05), so MBH over-smooths to a near-straight line.  As
# post-COVID recovery data arrive, that straight line tilts upward and
# revises the 2019 Q4 trend estimate — precisely the instability we want
# to avoid.  Fixing d = mad(hp_cycle) anchors the threshold to the right
# scale and keeps the backward revision near zero.
d_fixed <- stats::mad(hp_filter(us_gdp_vintage$gdp_log, freq = 4)$cycle)

# df = 10: avoids the extreme endpoint under-fit that df = 4 (default)
# produces.  With df = 4, the cycle at the 2019 Q4 endpoint is ~20 ppts;
# with df = 10 it is ~0.003 ppts (same as HP), removing the spurious
# upward drift that would otherwise accumulate as recovery data arrives.
# mstop = 1000 gives the additional boosting budget needed at df = 10.
fixed_df     <- 10L
fixed_mstop  <- 1000L

# Freeze the B-spline domain so every vintage uses the same basis — knot
# count and domain are set once from the full sample and never change.
fixed_knots  <- 50L                            # invariant functional space
fixed_bounds <- c(1L, T_max)                   # global B-spline domain

# Vintages: 2016 Q1 – 2022 Q4 (28 publication dates)
eval_dates   <- us_gdp_vintage[date >= "2016-01-01" & date <= "2022-10-01", date]
eval_indices <- which(us_gdp_vintage$date %in% eval_dates)

fan_list  <- vector("list", length(eval_indices))
back_list <- vector("list", length(eval_indices))

for (k in seq_along(eval_indices)) {
  i         <- eval_indices[k]
  y_current <- us_gdp_vintage$gdp_log[seq_len(i)]

  hp_res  <- hp_filter(y_current, freq = 4)

  # Single call per vintage — all parameters frozen outside the loop so
  # both the fan chart and the backward-revision series use the same basis.
  mbh_res <- mbh_filter(
    y_current,
    d              = d_fixed,
    knots          = fixed_knots,
    df             = fixed_df,
    mstop          = fixed_mstop,
    boundary.knots = fixed_bounds
  )

  n_cur    <- length(y_current)
  tail_idx <- max(1L, n_cur - 27L):n_cur   # 28-obs trailing window

  fan_list[[k]] <- data.table(
    vintage_date = us_gdp_vintage$date[i],
    obs_date     = us_gdp_vintage$date[tail_idx],
    hp_trend     = hp_res$trend[tail_idx],
    mbh_trend    = mbh_res$trend[tail_idx],
    gdp_log      = y_current[tail_idx]
  )

  if (i >= ref_idx) {
    back_list[[k]] <- data.table(
      vintage_date = us_gdp_vintage$date[i],
      hp_at_ref    = hp_res$trend[ref_idx],
      mbh_at_ref   = mbh_res$trend[ref_idx]
    )
  }
}

revisions_dt <- rbindlist(fan_list)
backward_dt  <- rbindlist(Filter(Negate(is.null), back_list))
```

---

## 3  The Role of `boundary.knots`

The MBH filter uses a B-spline basis $B(t)$ evaluated at the integer
time index $t = 1, 2, \ldots, n$.  By default `bbs()` sets the knot
domain to `range(time_idx) = c(1, n)`.  As $n$ grows by one, the domain
expands, the basis columns shift, and the estimated trend from vintage $k$
and vintage $k+1$ are **not numerically comparable** — differences partly
reflect a changed basis rather than new data.

Passing `boundary.knots = fixed_bounds` (= `c(1, T_max)`) to every
vintage call fixes the domain to the full-sample range.  All vintages
then share an identical B-spline column space, so differences between
consecutive trend estimates are attributable purely to the additional
data point, not to basis drift.

```{r boundary-knots-demo}
n_demo  <- 200L   # truncated sample
y_demo  <- us_gdp_vintage$gdp_log[seq_len(n_demo)]

res_free      <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = NULL)
res_anchored  <- mbh_filter(y_demo, knots = fixed_knots, df = fixed_df,
                             mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Extend by one observation and refit
y_demo_p1 <- us_gdp_vintage$gdp_log[seq_len(n_demo + 1L)]

res_free_p1     <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = NULL)
res_anchored_p1 <- mbh_filter(y_demo_p1, knots = fixed_knots, df = fixed_df,
                               mstop = fixed_mstop, boundary.knots = fixed_bounds)

# Revision at the final shared observation (position n_demo)
rev_free     <- abs(res_free_p1$trend[n_demo]     - res_free$trend[n_demo])
rev_anchored <- abs(res_anchored_p1$trend[n_demo] - res_anchored$trend[n_demo])

cat(sprintf(
  "Revision at obs %d after adding one data point:\n  free domain   : %.6f\n  anchored domain: %.6f\n",
  n_demo, rev_free, rev_anchored
))
```

The anchored revision is smaller because a fixed `boundary.knots` removes
a numerical artefact that would otherwise masquerade as genuine trend
movement.

---

## 4  Vintage Fan Chart

Each semi-transparent line below is one vintage estimate.  A **wide fan**
(many non-overlapping lines) means the filter is revising its trend view
strongly as new data arrive — a sign of end-point instability.  A **tight
bundle** means the filter is stable in real time.

```{r fan-chart}
# Show only the shared overlap window (2018 Q1 onward) where every one of
# the 28 vintages contributes data — this eliminates the staircase / accordion
# artefact that arises when staggered trailing windows are plotted together.
fan_shared <- revisions_dt[obs_date >= as.Date("2018-01-01")]

p1 <- ggplot(fan_shared, aes(x = obs_date)) +
  geom_line(aes(y = hp_trend,  group = vintage_date),
            colour = "#0072B2", alpha = 0.4, linewidth = 0.6) +
  geom_line(aes(y = mbh_trend, group = vintage_date),
            colour = "#E69F00", alpha = 0.4, linewidth = 0.6) +
  geom_line(
    data = us_gdp_vintage[date >= as.Date("2018-01-01") & date <= as.Date("2022-10-01")],
    aes(x = date, y = gdp_log),
    colour = "black", linewidth = 0.8, linetype = "dashed"
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.08, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  labs(
    title    = "Vintage Fan Chart: HP (blue) vs MBH (orange)",
    subtitle = "Each line = one vintage estimate (2018–2022 overlap window). Dashed = observed.",
    x        = NULL,
    y        = "Log Real GDP"
  ) +
  theme_minimal(base_size = 12)

print(p1)
```

The blue HP fan opens visibly around the 2020 COVID quarters — the filter
revises its trend estimate substantially as new post-COVID data become
available.  The orange MBH bundle remains tight: the Huber loss prevents
the COVID shock from distorting the trend, so successive vintages produce
nearly identical estimates.

---

## 5  The Backward Revision Test

The most direct proof of end-point instability is not the *current*
trend estimate but **how a past estimate gets rewritten** as new data
arrive.  Pick a fixed pre-COVID reference date — 2019 Q4 — and ask:
*how does each filter's trend estimate at that date change as we add
2020–2022 data?*

For HP, the COVID shock at the end of the sample bends the smoother
backward: 2019 Q4's trend is revised *downward* when COVID arrives
(HP interprets the collapse as trend), then partially corrected as
recovery data is added.  For MBH, the Huber loss caps the influence of
the outlier, so earlier trend estimates barely move.

```{r backward-revision}
# backward_dt was built in the expanding-window loop above.
# MBH used df = 10 (correct endpoint estimate), knots = 50, and
# boundary.knots = c(1, T_max) (frozen basis) — so all vintages are
# numerically comparable and the baseline trend is unbiased.
back_dt <- backward_dt[order(vintage_date)]
setnames(back_dt, c("hp_at_ref", "mbh_at_ref"), c("hp_trend", "mbh_trend"))

# Normalise to the base vintage (2019 Q4 = first vintage that includes ref_date)
base_hp  <- back_dt[vintage_date == ref_date, hp_trend]
base_mbh <- back_dt[vintage_date == ref_date, mbh_trend]

back_dt[, hp_revision  := (hp_trend  - base_hp)  * 100]   # × 100 = ppts log GDP
back_dt[, mbh_revision := (mbh_trend - base_mbh) * 100]

p2 <- ggplot(back_dt, aes(x = vintage_date)) +
  geom_hline(yintercept = 0, linetype = "dashed", colour = "grey60") +
  geom_line(aes(y = hp_revision),  colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = hp_revision), colour = "#0072B2", size = 2.5) +
  geom_line(aes(y = mbh_revision),  colour = "#E69F00", linewidth = 1) +
  geom_point(aes(y = mbh_revision), colour = "#E69F00", size = 2.5) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID\nshock", vjust = 1.4, size = 3.2, colour = "firebrick") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$hp_revision, 1), label = "HP",
           hjust = -0.3, colour = "#0072B2", fontface = "bold") +
  annotate("text", x = max(back_dt$vintage_date),
           y = tail(back_dt$mbh_revision, 1), label = "MBH",
           hjust = -0.3, colour = "#E69F00", fontface = "bold") +
  scale_x_date(
    date_breaks = "6 months",
    labels      = function(x) paste0(format(x, "%Y"), " ", quarters(x)),
    expand      = expansion(add = c(30, 90))
  ) +
  labs(
    title    = "Backward Revision at 2019 Q4 as New Data Arrive",
    subtitle = "How does each filter re-estimate the pre-COVID trend as COVID data are published?",
    x        = "Vintage (publication date)",
    y        = "Revision vs 2019 Q4 baseline (log pts x100)"
  ) +
  theme_minimal(base_size = 11) +
  theme(axis.title = element_text(size = 9))

print(p2)
```

HP (blue) dips sharply when 2020 Q2 data arrive — the filter
re-attributes pre-COVID growth as "trend decline" — then partially
recovers as post-COVID data accumulate.  MBH (orange) dips roughly half
as much: the Huber loss limits how much the COVID observation can drag the
estimated pre-COVID trend downward, so the backward revision is
substantially smaller than HP's throughout the simulation window.

```{r spread-table}
knitr::kable(
  back_dt[, .(
    Vintage              = format(vintage_date),
    `HP trend at 2019Q4`  = round(hp_trend,  5),
    `MBH trend at 2019Q4` = round(mbh_trend, 5),
    `HP revision (ppts)`  = round(hp_revision,  3),
    `MBH revision (ppts)` = round(mbh_revision, 3)
  )],
  caption = "Backward revision of 2019 Q4 trend estimate across vintages"
)
```

---

## Summary

| Property | HP Filter | MBH Filter |
|---|---|---|
| End-point leverage | High (boundary asymmetry) | Low (Huber down-weighting) |
| Backward revision of pre-COVID trend | Large (dips on shock) | Near-zero |
| COVID trend distortion | Severe | Minimal |
| Basis stability across vintages | N/A | Requires `boundary.knots` |

For real-time monitoring applications — nowcasting output gaps, setting
automatic stabilisers, informing monetary policy decisions — the MBH
filter's robustness substantially reduces the risk that a preliminary
trend estimate is later revised beyond recognition.

---
title: "Uncertainty Bands via Block Bootstrap"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Uncertainty Bands via Block Bootstrap}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4.5,
  out.width  = "100%"
)
```

```{r setup, message=FALSE}
library(MacroFilters)
library(ggplot2)
data("fr_gdp", package = "MacroFilters")

# France real GDP (log level) as a quarterly ts; includes the 2020 Q2 COVID shock
d0 <- fr_gdp$date[1]
fr <- ts(
  fr_gdp$gdp_log,
  start     = c(as.integer(format(d0, "%Y")),
                (as.integer(format(d0, "%m")) - 1L) %/% 3L + 1L),
  frequency = 4
)
```

---

## 1. Why quantify trend uncertainty?

A point estimate of the trend is only half the story. In real time the trend is
most uncertain exactly where it matters most — at the **end of the sample**,
where the smoother has no future observations to lean on. Every filter in
**MacroFilters** can attach a 95% confidence band to its trend through a single
argument, `boot_iter`.

```{r basic-band}
fit <- mbh_filter(fr, boot_iter = 50L)   # mstop defaults to 500
str(fit[c("trend_lower", "trend_upper")], max.level = 1)
```

When `boot_iter > 0` the returned object gains `$trend_lower` and
`$trend_upper`; otherwise they are absent. Plotting is automatic — `autoplot()`
draws the ribbon whenever the bands are present:

```{r basic-plot}
autoplot(fit)
```

The trend cuts almost straight through the 2020 COVID collapse: the Huber loss
treats the shock as an outlier rather than bending the trend toward it.

---

## 2. The mechanics

The engine is a **Circular Block Bootstrap** (Politis & Romano, 1992) of the
filter's pseudo-residuals (the cycle):

1. Resample the cycle in contiguous **blocks** that wrap around the series end,
   preserving short-run autocorrelation while giving every observation equal
   weight (no end-point under-representation).
2. Rebuild a synthetic series `trend + resampled cycle` and **refit the same
   filter** to it. Each refit uses the *same estimator* as the base fit (same
   `mstop` for MBH, same iteration count for bHP), so the band width is not
   biased.
3. The band is the **normal approximation** `trend ± 1.96 * sd(bootstrap trends)`,
   centred on the point estimate. The standard deviation is used instead of raw
   2.5%/97.5% percentiles because it is smooth and stable at a practical
   `boot_iter` (percentiles need hundreds of replicates to avoid jitter).

Two knobs control it:

- **`boot_iter`** — number of bootstrap replicates (cost grows linearly).
- **`block_size`** — block length; `"auto"` (default) uses `2 * frequency`
  (two cycles), capped at `length(x) / 3` to keep at least three blocks.

```{r block-size}
# Quarterly data -> auto block size = 2 * 4 = 8
fit_b <- mbh_filter(fr, boot_iter = 50L, block_size = 8L)
```

---

## 3. Bands for every filter

The same machinery is available on all four filters. Note that MBH keeps the
default `mstop = 500`: for long log-level series, reducing `mstop` collapses the
trend (the Huber gradient is capped from the first iteration, so the trend never
climbs its full range).

```{r all-filters, fig.height=3.6}
autoplot(hp_filter(fr,       boot_iter = 50L))
autoplot(bhp_filter(fr,      boot_iter = 50L))
autoplot(hamilton_filter(fr, boot_iter = 50L))
autoplot(mbh_filter(fr,      boot_iter = 50L))
```

### Reading the end-point fan

For all filters the band **widens toward the edges** — typically two to three
times the mid-sample width. This is honest, not a bug: estimating the trend at
`t = n` with no future data is genuinely far more uncertain than in the
interior. It is precisely the *end-point problem* these filters are designed to
expose.

### A note on the Hamilton band

The Hamilton filter is a regression on lags of the series, so its first
`h + p - 1` observations have no fitted trend (they appear as `NA`, leaving a
gap at the start of the plot). Its bootstrap is **conditional on those initial
observations**: the lead-in is held fixed across replicates. As a result the
band is narrow at the start of the valid window — where the predictors are the
frozen lead-in and only the regression coefficients vary — and widens forward as
the predictors themselves become resampled quantities.

---

## 4. References

- Politis, D.N. and Romano, J.P. (1992). A circular block-resampling procedure
  for stationary data. In *Exploring the Limits of the Bootstrap*, 263–270.
- Künsch, H.R. (1989). The jackknife and the bootstrap for general stationary
  observations. *Annals of Statistics*, 17(3), 1217–1241.

---
title: "Hyperparameter Tuning for the MacroBoost Hybrid Filter"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Hyperparameter Tuning for the MacroBoost Hybrid Filter}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")
```

## 1  Anatomy of the MBH Trifecta

`mbh_filter()` exposes three knobs that jointly govern fit quality,
robustness, and computation time.

| Parameter | Default | Role |
|---|---|---|
| `mstop` | `500` | Gradient-descent iteration budget |
| `nu` | `0.1` | Per-step shrinkage (learning rate) |
| `knots` | `min(max(20, floor(n/2)), 250)` | P-spline interior knot count (capped at 250) |

### 1.1  `mstop` — iteration budget

Each boosting step adds a small, smooth update to the current trend
estimate.  More iterations allow finer approximation of the optimal
solution, at the cost of linear wall time.  The default `mstop = 500`
is calibrated to produce HP-comparable flexibility on quarterly macro
series (~100–400 observations).

### 1.2  `nu` — shrinkage

`nu` scales the gradient contribution of each step.  Lower `nu` requires
more iterations to reach the same fit quality; higher `nu` converges
faster but may overshoot.  The key identity is:

$$\text{effective update} \approx \nu \times \text{mstop} \times \text{step size}$$

A practical equivalence: `(mstop = 500, nu = 0.1)` and
`(mstop = 1000, nu = 0.05)` produce nearly identical trends.

```{r nu-equiv}
y <- us_gdp_vintage$gdp_log

res_default <- mbh_filter(y, mstop = 500,  nu = 0.10)
res_equiv   <- mbh_filter(y, mstop = 1000, nu = 0.05)

max_diff <- max(abs(res_default$trend - res_equiv$trend))
cat(sprintf("Max trend difference (mstop×nu equivalence): %.2e\n", max_diff))
```

### 1.3  `knots` — spline flexibility

Knots control how many local basis functions the P-spline uses to
represent the trend shape.  The default heuristic `min(max(20, floor(n/2)), 250)`
provides roughly one knot per two observations (capped at 250 for long
series) — high density relative to classical smoothing spline
recommendations — because Huber loss (not spline regularity) acts as the
primary smoothness constraint.

Too few knots force an overly rigid polynomial-like trend; too many knots
with low `mstop` may leave some basis functions unfitted.

---

## 2  Auto-calibration of Huber Delta `d`

The Huber loss function

$$L_\delta(r) = \begin{cases} \tfrac{1}{2}r^2 & |r| \le \delta \\ \delta\,|r| - \tfrac{1}{2}\delta^2 & |r| > \delta \end{cases}$$

transitions between $L_2$ (quadratic) loss for small residuals and $L_1$
(absolute) loss for large ones.  The threshold $\delta$ determines which
residuals are treated as outliers.

When `d = NULL`, the threshold is set automatically via the **MAD of
first differences**:

$$\hat{d} = \text{MAD}(\Delta y) = \frac{\text{median}(|\Delta y - \text{median}(\Delta y)|)}{0.6745}$$

The $0.6745$ scale factor makes MAD a consistent estimator of the standard
deviation under normality.

### 2.1  Scale invariance

If $y \to a \cdot y$, then $\Delta y \to a \cdot \Delta y$ and
$\hat{d} \to a \cdot \hat{d}$.  The Huber threshold scales exactly with the
data amplitude, so the filter behaves identically regardless of the
measurement units.

```{r scale-invariance}
y_level <- us_gdp_vintage$gdp_real   # billions USD (~20 000 scale)
y_log   <- us_gdp_vintage$gdp_log    # natural log (~10 scale)

d_level <- stats::mad(diff(y_level))
d_log   <- stats::mad(diff(y_log))

cat(sprintf("d (level series) : %.4f\n", d_level))
cat(sprintf("d (log series)   : %.6f\n", d_log))
cat(sprintf("Ratio d_level / mean(level): %.6f\n", d_level / mean(y_level)))
cat(sprintf("Ratio d_log   / mean(log)  : %.6f\n", d_log   / mean(y_log)))
```

### 2.2  Scale-mismatch warning for log-level input

For log-level GDP, `diff(y)` returns quarterly growth rates whose typical
scale is 0.004–0.010.  The output gap (the cycle the filter must explain)
operates on a much larger scale, typically 0.01–0.05.  Using
`d = mad(diff(y))` therefore sets the Huber threshold **too tight**:
normal business-cycle oscillations are mis-classified as outliers, their
gradient contributions are truncated, and the trend becomes a near-straight
line.

The recommended override is:

```r
d_cycle <- mad(hp_filter(y_log)$cycle)   # set d on the residual scale
res     <- mbh_filter(y_log, d = d_cycle)
```

---

## 3  Overriding `d` for High-Volatility Series

Quarterly GDP growth rates (`diff(log(GDP))`) are roughly 40× more
volatile relative to trend than log levels.  The COVID collapse of 2020 Q2
($\approx -9\%$ q-o-q) represents an extreme outlier even by growth-rate
standards.  This makes growth rates an ideal stress test for `d` sensitivity.

```{r d-sensitivity}
y_growth <- diff(us_gdp_vintage$gdp_log)   # quarterly log-differences

res_auto   <- mbh_filter(y_growth)
res_strict <- mbh_filter(y_growth, d = 0.005)
res_lenient <- mbh_filter(y_growth, d = 0.02)

cat(sprintf("Auto d = %.6f\n", res_auto$meta$d))
```

```{r d-sensitivity-plot}
dt_growth <- data.table(
  t        = us_gdp_vintage$date[-1],
  observed = y_growth,
  auto     = res_auto$trend,
  strict   = res_strict$trend,
  lenient  = res_lenient$trend
)

dt_long <- melt(dt_growth,
                id.vars      = "t",
                measure.vars = c("auto", "strict", "lenient"),
                variable.name = "delta",
                value.name    = "trend")

# Human-readable labels
auto_label <- sprintf("Auto (d=%.4f)", res_auto$meta$d)
# data.table::melt() returns variable.name as factor; fcase() returns character.
# Assigning character to a factor column via := raises a type mismatch error,
# so coerce to character first.
dt_long[, delta := as.character(delta)]
dt_long[, delta := fcase(
  delta == "auto",    auto_label,
  delta == "strict",  "Strict (d=0.005)",
  delta == "lenient", "Lenient (d=0.020)"
)]

colour_vals <- c("#0072B2", "#009E73", "#E69F00")
names(colour_vals) <- c("Strict (d=0.005)", auto_label, "Lenient (d=0.020)")

p_d <- ggplot() +
  geom_line(
    data = dt_growth,
    aes(x = t, y = observed),
    colour = "grey70", linewidth = 0.5
  ) +
  geom_line(
    data = dt_long,
    aes(x = t, y = trend, colour = delta),
    linewidth = 0.9
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID Q2\n-9% q-o-q", vjust = 1.4,
           size = 3.2, colour = "firebrick") +
  scale_colour_manual(values = colour_vals) +
  labs(
    title    = "MBH Trend Sensitivity to Huber Delta d",
    subtitle = "Data: US quarterly GDP growth rates (log-diff)",
    x        = NULL, y = "Log-difference", colour = "d setting"
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "top")

print(p_d)
```

**Interpretation**

- **Strict `d = 0.005`** (blue): Huber threshold is tight; even modest
  growth rate swings are down-weighted.  The trend is nearly flat,
  absorbing almost no cyclical signal.
- **Auto `d`** (green): threshold is calibrated to normal volatility.
  The COVID spike is substantially down-weighted but ordinary fluctuations
  are fitted.
- **Lenient `d = 0.020`** (orange): threshold is loose; the filter
  behaves close to $L_2$ boosting and the trend responds to the COVID shock.

---

## 4  Computational Trade-off Benchmark

```{r benchmark, cache=TRUE}
y          <- us_gdp_vintage$gdp_log
mstop_grid <- seq(100L, 1000L, by = 100L)   # 10 evenly-spaced points
n_rep      <- 5L                            # replicates per point

# Single-shot timings are dominated by GC pauses and OS scheduling, which
# produce spurious spikes. We warm up once, then time several runs with a
# clean heap (gc()) before each and report the MEDIAN, recovering the
# monotone wall-time vs mstop relationship.
bench_dt <- rbindlist(lapply(mstop_grid, function(m) {
  res  <- suppressMessages(mbh_filter(y, mstop = m))   # warm-up + cycle SD
  reps <- vapply(seq_len(n_rep), function(i) {
    gc(verbose = FALSE)
    t0 <- proc.time()
    suppressMessages(mbh_filter(y, mstop = m))
    (proc.time() - t0)[["elapsed"]]
  }, numeric(1))
  data.table(
    mstop       = m,
    elapsed_sec = round(median(reps), 3),
    cycle_sd    = round(sd(res$cycle), 6)
  )
}))

knitr::kable(
  bench_dt,
  col.names = c("mstop", "Wall time (s)", "Cycle SD"),
  caption   = sprintf("MBH computational benchmark — US log GDP (%d obs)", length(y))
)
```

```{r benchmark-plot}
# Dual-axis layout: wall time (left) + cycle_sd convergence (right)
# Use a secondary-axis trick by normalising cycle_sd to the time scale
time_range  <- range(bench_dt$elapsed_sec)
sd_range    <- range(bench_dt$cycle_sd)
# Guard against division by zero if cycle_sd converges to a flat line
if (diff(sd_range)   < 1e-10) sd_range   <- sd_range   + c(-1e-5,  1e-5)
if (diff(time_range) < 1e-10) time_range <- time_range + c(-1e-5,  1e-5)
sd_to_time  <- function(x) (x - sd_range[1]) / diff(sd_range) * diff(time_range) + time_range[1]
time_to_sd  <- function(x) (x - time_range[1]) / diff(time_range) * diff(sd_range) + sd_range[1]

p_bench <- ggplot(bench_dt, aes(x = mstop)) +
  geom_line(aes(y = elapsed_sec), colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = elapsed_sec), colour = "#CC0000", size = 3) +
  geom_line(aes(y = sd_to_time(cycle_sd)),
            colour = "#E69F00", linewidth = 0.9, linetype = "dashed") +
  geom_point(aes(y = sd_to_time(cycle_sd)),
             colour = "#E69F00", size = 2.5) +
  scale_x_continuous(breaks = mstop_grid) +
  scale_y_continuous(
    name     = "Wall time (s)  [blue / red points]",
    sec.axis = sec_axis(~ time_to_sd(.), name = "Cycle SD  [orange dashed]",
                        labels = scales::label_number(accuracy = 0.0001))
  ) +
  labs(
    title    = "Wall Time vs Boosting Iterations",
    subtitle = sprintf("US Real GDP log level (%d obs). Cycle SD plateaus well before mstop = 500.", length(y)),
    x        = "mstop"
  ) +
  theme_minimal(base_size = 12) +
  theme(
    axis.title.y.left  = element_text(colour = "#0072B2"),
    axis.title.y.right = element_text(colour = "#E69F00")
  )

print(p_bench)
```

### Practical guidance

| Use case | Recommended settings |
|---|---|
| Interactive / exploratory | `mstop = 100–200` |
| Publication-quality output | `mstop = 500` (default) |
| Long daily series (n > 5 000) | `mstop = 50, nu = 0.3` |
| Cross-country batch estimation | `mstop = 200, nu = 0.15` |

Gains in cycle standard deviation (a proxy for fit quality) diminish
rapidly above `mstop = 200` for a typical macro series.  The default
`mstop = 500` provides a comfortable safety margin without being
prohibitively slow.

---

## 5  Summary

```{r summary-table, echo=FALSE}
summary_tbl <- data.table(
  Parameter = c("`mstop`", "`nu`", "`knots`", "`d`"),
  Default   = c("500", "0.1", "`min(max(20, n/2), 250)`", "auto via MAD"),
  `When to increase` = c(
    "Publication accuracy required",
    "Very long series; computational budget tight",
    "Highly nonlinear trend",
    "Series has frequent large spikes"
  ),
  `When to decrease` = c(
    "Exploratory / fast iteration",
    "Stability preferred over speed",
    "Short series or near-linear trend",
    "Series is log-level (use `mad(hp$cycle)` instead)"
  )
)
knitr::kable(summary_tbl, caption = "MBH hyperparameter quick-reference")
```

---
title: "Hyperparameter Tuning for the MacroBoost Hybrid Filter"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Hyperparameter Tuning for the MacroBoost Hybrid Filter}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE, comment = "#>",
  fig.width = 8, fig.height = 4.5, out.width = "100%"
)
library(MacroFilters)
library(data.table)
library(ggplot2)
data("us_gdp_vintage", package = "MacroFilters")
```

## 1  Anatomy of the MBH Trifecta

`mbh_filter()` exposes three knobs that jointly govern fit quality,
robustness, and computation time.

| Parameter | Default | Role |
|---|---|---|
| `mstop` | `500` | Gradient-descent iteration budget |
| `nu` | `0.1` | Per-step shrinkage (learning rate) |
| `knots` | `max(20, floor(n/2))` | P-spline interior knot count |

### 1.1  `mstop` — iteration budget

Each boosting step adds a small, smooth update to the current trend
estimate.  More iterations allow finer approximation of the optimal
solution, at the cost of linear wall time.  The default `mstop = 500`
is calibrated to produce HP-comparable flexibility on quarterly macro
series (~100–400 observations).

### 1.2  `nu` — shrinkage

`nu` scales the gradient contribution of each step.  Lower `nu` requires
more iterations to reach the same fit quality; higher `nu` converges
faster but may overshoot.  The key identity is:

$$\text{effective update} \approx \nu \times \text{mstop} \times \text{step size}$$

A practical equivalence: `(mstop = 500, nu = 0.1)` and
`(mstop = 1000, nu = 0.05)` produce nearly identical trends.

```{r nu-equiv}
y <- us_gdp_vintage$gdp_log

res_default <- mbh_filter(y, mstop = 500,  nu = 0.10)
res_equiv   <- mbh_filter(y, mstop = 1000, nu = 0.05)

max_diff <- max(abs(res_default$trend - res_equiv$trend))
cat(sprintf("Max trend difference (mstop×nu equivalence): %.2e\n", max_diff))
```

### 1.3  `knots` — spline flexibility

Knots control how many local basis functions the P-spline uses to
represent the trend shape.  The default heuristic `max(20, floor(n/2))`
provides roughly one knot per two observations — high density relative to
classical smoothing spline recommendations — because Huber loss (not spline
regularity) acts as the primary smoothness constraint.

Too few knots force an overly rigid polynomial-like trend; too many knots
with low `mstop` may leave some basis functions unfitted.

---

## 2  Auto-calibration of Huber Delta `d`

The Huber loss function

$$L_\delta(r) = \begin{cases} \tfrac{1}{2}r^2 & |r| \le \delta \\ \delta\,|r| - \tfrac{1}{2}\delta^2 & |r| > \delta \end{cases}$$

transitions between $L_2$ (quadratic) loss for small residuals and $L_1$
(absolute) loss for large ones.  The threshold $\delta$ determines which
residuals are treated as outliers.

When `d = NULL`, the threshold is set automatically via the **MAD of
first differences**:

$$\hat{d} = \text{MAD}(\Delta y) = \frac{\text{median}(|\Delta y - \text{median}(\Delta y)|)}{0.6745}$$

The $0.6745$ scale factor makes MAD a consistent estimator of the standard
deviation under normality.

### 2.1  Scale invariance

If $y \to a \cdot y$, then $\Delta y \to a \cdot \Delta y$ and
$\hat{d} \to a \cdot \hat{d}$.  The Huber threshold scales exactly with the
data amplitude, so the filter behaves identically regardless of the
measurement units.

```{r scale-invariance}
y_level <- us_gdp_vintage$gdp_real   # billions USD (~20 000 scale)
y_log   <- us_gdp_vintage$gdp_log    # natural log (~10 scale)

d_level <- stats::mad(diff(y_level))
d_log   <- stats::mad(diff(y_log))

cat(sprintf("d (level series) : %.4f\n", d_level))
cat(sprintf("d (log series)   : %.6f\n", d_log))
cat(sprintf("Ratio d_level / mean(level): %.6f\n", d_level / mean(y_level)))
cat(sprintf("Ratio d_log   / mean(log)  : %.6f\n", d_log   / mean(y_log)))
```

### 2.2  Scale-mismatch warning for log-level input

For log-level GDP, `diff(y)` returns quarterly growth rates whose typical
scale is 0.004–0.010.  The output gap (the cycle the filter must explain)
operates on a much larger scale, typically 0.01–0.05.  Using
`d = mad(diff(y))` therefore sets the Huber threshold **too tight**:
normal business-cycle oscillations are mis-classified as outliers, their
gradient contributions are truncated, and the trend becomes a near-straight
line.

The recommended override is:

```r
d_cycle <- mad(hp_filter(y_log)$cycle)   # set d on the residual scale
res     <- mbh_filter(y_log, d = d_cycle)
```

---

## 3  Overriding `d` for High-Volatility Series

Quarterly GDP growth rates (`diff(log(GDP))`) are roughly 40× more
volatile relative to trend than log levels.  The COVID collapse of 2020 Q2
($\approx -9\%$ q-o-q) represents an extreme outlier even by growth-rate
standards.  This makes growth rates an ideal stress test for `d` sensitivity.

```{r d-sensitivity}
y_growth <- diff(us_gdp_vintage$gdp_log)   # quarterly log-differences

res_auto   <- mbh_filter(y_growth)
res_strict <- mbh_filter(y_growth, d = 0.005)
res_lenient <- mbh_filter(y_growth, d = 0.02)

cat(sprintf("Auto d = %.6f\n", res_auto$meta$d))
```

```{r d-sensitivity-plot}
dt_growth <- data.table(
  t        = us_gdp_vintage$date[-1],
  observed = y_growth,
  auto     = res_auto$trend,
  strict   = res_strict$trend,
  lenient  = res_lenient$trend
)

dt_long <- melt(dt_growth,
                id.vars      = "t",
                measure.vars = c("auto", "strict", "lenient"),
                variable.name = "delta",
                value.name    = "trend")

# Human-readable labels
auto_label <- sprintf("Auto (d=%.4f)", res_auto$meta$d)
# data.table::melt() returns variable.name as factor; fcase() returns character.
# Assigning character to a factor column via := raises a type mismatch error,
# so coerce to character first.
dt_long[, delta := as.character(delta)]
dt_long[, delta := fcase(
  delta == "auto",    auto_label,
  delta == "strict",  "Strict (d=0.005)",
  delta == "lenient", "Lenient (d=0.020)"
)]

colour_vals <- c("#0072B2", "#009E73", "#E69F00")
names(colour_vals) <- c("Strict (d=0.005)", auto_label, "Lenient (d=0.020)")

p_d <- ggplot() +
  geom_line(
    data = dt_growth,
    aes(x = t, y = observed),
    colour = "grey70", linewidth = 0.5
  ) +
  geom_line(
    data = dt_long,
    aes(x = t, y = trend, colour = delta),
    linewidth = 0.9
  ) +
  annotate("rect",
           xmin = as.Date("2020-01-01"), xmax = as.Date("2020-10-01"),
           ymin = -Inf, ymax = Inf, alpha = 0.1, fill = "firebrick") +
  annotate("text", x = as.Date("2020-04-01"), y = Inf,
           label = "COVID Q2\n-9% q-o-q", vjust = 1.4,
           size = 3.2, colour = "firebrick") +
  scale_colour_manual(values = colour_vals) +
  labs(
    title    = "MBH Trend Sensitivity to Huber Delta d",
    subtitle = "Data: US quarterly GDP growth rates (log-diff)",
    x        = NULL, y = "Log-difference", colour = "d setting"
  ) +
  theme_minimal(base_size = 12) +
  theme(legend.position = "top")

print(p_d)
```

**Interpretation**

- **Strict `d = 0.005`** (blue): Huber threshold is tight; even modest
  growth rate swings are down-weighted.  The trend is nearly flat,
  absorbing almost no cyclical signal.
- **Auto `d`** (green): threshold is calibrated to normal volatility.
  The COVID spike is substantially down-weighted but ordinary fluctuations
  are fitted.
- **Lenient `d = 0.020`** (orange): threshold is loose; the filter
  behaves close to $L_2$ boosting and the trend responds to the COVID shock.

---

## 4  Computational Trade-off Benchmark

```{r benchmark, cache=TRUE}
y          <- us_gdp_vintage$gdp_log
mstop_grid <- seq(100L, 1000L, by = 100L)   # 10 evenly-spaced points

bench_dt <- rbindlist(lapply(mstop_grid, function(m) {
  t0       <- proc.time()
  res      <- mbh_filter(y, mstop = m)
  elapsed  <- (proc.time() - t0)[["elapsed"]]
  cycle_sd <- sd(res$cycle)
  data.table(
    mstop       = m,
    elapsed_sec = round(elapsed, 3),
    cycle_sd    = round(cycle_sd, 6)
  )
}))

knitr::kable(
  bench_dt,
  col.names = c("mstop", "Wall time (s)", "Cycle SD"),
  caption   = "MBH computational benchmark — US log GDP (316 obs)"
)
```

```{r benchmark-plot}
# Dual-axis layout: wall time (left) + cycle_sd convergence (right)
# Use a secondary-axis trick by normalising cycle_sd to the time scale
time_range  <- range(bench_dt$elapsed_sec)
sd_range    <- range(bench_dt$cycle_sd)
# Guard against division by zero if cycle_sd converges to a flat line
if (diff(sd_range)   < 1e-10) sd_range   <- sd_range   + c(-1e-5,  1e-5)
if (diff(time_range) < 1e-10) time_range <- time_range + c(-1e-5,  1e-5)
sd_to_time  <- function(x) (x - sd_range[1]) / diff(sd_range) * diff(time_range) + time_range[1]
time_to_sd  <- function(x) (x - time_range[1]) / diff(time_range) * diff(sd_range) + sd_range[1]

p_bench <- ggplot(bench_dt, aes(x = mstop)) +
  geom_line(aes(y = elapsed_sec), colour = "#0072B2", linewidth = 1) +
  geom_point(aes(y = elapsed_sec), colour = "#CC0000", size = 3) +
  geom_line(aes(y = sd_to_time(cycle_sd)),
            colour = "#E69F00", linewidth = 0.9, linetype = "dashed") +
  geom_point(aes(y = sd_to_time(cycle_sd)),
             colour = "#E69F00", size = 2.5) +
  scale_x_continuous(breaks = mstop_grid) +
  scale_y_continuous(
    name     = "Wall time (s)  [blue / red points]",
    sec.axis = sec_axis(~ time_to_sd(.), name = "Cycle SD  [orange dashed]",
                        labels = scales::label_number(accuracy = 0.0001))
  ) +
  labs(
    title    = "Wall Time vs Boosting Iterations",
    subtitle = "US Real GDP log level (316 obs). Cycle SD plateaus well before mstop = 500.",
    x        = "mstop"
  ) +
  theme_minimal(base_size = 12) +
  theme(
    axis.title.y.left  = element_text(colour = "#0072B2"),
    axis.title.y.right = element_text(colour = "#E69F00")
  )

print(p_bench)
```

### Practical guidance

| Use case | Recommended settings |
|---|---|
| Interactive / exploratory | `mstop = 100–200` |
| Publication-quality output | `mstop = 500` (default) |
| Long daily series (n > 5 000) | `mstop = 50, nu = 0.3` |
| Cross-country batch estimation | `mstop = 200, nu = 0.15` |

Gains in cycle standard deviation (a proxy for fit quality) diminish
rapidly above `mstop = 200` for a typical macro series.  The default
`mstop = 500` provides a comfortable safety margin without being
prohibitively slow.

---

## 5  Summary

```{r summary-table, echo=FALSE}
summary_tbl <- data.table(
  Parameter = c("`mstop`", "`nu`", "`knots`", "`d`"),
  Default   = c("500", "0.1", "`max(20, n/2)`", "auto via MAD"),
  `When to increase` = c(
    "Publication accuracy required",
    "Very long series; computational budget tight",
    "Highly nonlinear trend",
    "Series has frequent large spikes"
  ),
  `When to decrease` = c(
    "Exploratory / fast iteration",
    "Stability preferred over speed",
    "Short series or near-linear trend",
    "Series is log-level (use `mad(hp$cycle)` instead)"
  )
)
knitr::kable(summary_tbl, caption = "MBH hyperparameter quick-reference")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/02_algo_hp.R
\name{hp_filter}
\alias{hp_filter}
\title{Hodrick-Prescott Filter (Sparse Matrix Implementation)}
\usage{
hp_filter(x, lambda = NULL, freq = NULL)
}
\arguments{
\item{x}{Numeric vector, \code{ts}, \code{xts}, or \code{zoo} object.}

\item{lambda}{Smoothing parameter.
If \code{NULL} (default), it is auto-detected using the Ravn-Uhlig rule
(\code{6.25 * freq^4}).}

\item{freq}{Numeric frequency override (1 = annual, 4 = quarterly,
12 = monthly). Used only when \code{lambda} is \code{NULL} and the frequency cannot
be inferred from \code{x}.}
}
\value{
A \code{macrofilter} object with \code{trend}, \code{cycle}, \code{data}, and \code{meta}
components.
}
\description{
Decomposes a time series into trend and cycle components by solving the
HP penalized least-squares problem using a sparse Cholesky factorization.
This avoids the dense O(n^3) inversion used by other implementations and
scales linearly in the number of observations.
}
\details{
The HP filter minimises
\deqn{\sum (y_t - \tau_t)^2 + \lambda \sum (\Delta^2 \tau_t)^2}
which admits the closed-form solution
\deqn{(I + \lambda D'D)\,\tau = y}
where \eqn{D} is the second-difference operator.

The implementation builds \eqn{D} as a banded sparse matrix
(\code{\link[Matrix:bandSparse]{Matrix::bandSparse()}}) and solves the symmetric positive-definite system
with a sparse Cholesky decomposition (\code{\link[Matrix:solve-methods]{Matrix::solve()}}).

When \code{lambda} is not supplied the Ravn-Uhlig (2002) rule is applied:
\code{lambda = 6.25 * freq^4}, yielding 6.25 (annual), 1600 (quarterly), and
129 600 (monthly).
}
\examples{
# Quarterly GDP-like series
y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
result <- hp_filter(y)
print(result)
}
\references{
Hodrick, R.J. and Prescott, E.C. (1997). Postwar U.S. Business Cycles: An
Empirical Investigation. \emph{Journal of Money, Credit and Banking}, 29(1),
1--16.

Ravn, M.O. and Uhlig, H. (2002). On Adjusting the Hodrick-Prescott Filter
for the Frequency of Observations. \emph{Review of Economics and Statistics},
84(2), 371--376.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{us_gdp_vintage}
\alias{us_gdp_vintage}
\title{US Real GDP — FRED Vintage}
\format{
A \code{data.table} with one row per quarter and three columns:
\describe{
\item{\code{date}}{\code{Date}. Quarter start date (e.g. \code{1947-01-01} = 1947 Q1).}
\item{\code{gdp_real}}{\code{numeric}. Real GDP level, billions of chained 2017 USD.}
\item{\code{gdp_log}}{\code{numeric}. Natural logarithm of \code{gdp_real}, pre-computed
for convenience.}
}
}
\source{
Federal Reserve Bank of St. Louis — FRED Economic Data,
series GDPC1. Downloaded via the public CSV endpoint
\verb{https://fred.stlouisfed.org/graph/fredgraph.csv?id=GDPC1}.
See \code{data-raw/us_gdp_vintage.R} for the reproducible download script.
}
\usage{
us_gdp_vintage
}
\description{
Quarterly US Real Gross Domestic Product from the Federal Reserve Bank of
St. Louis (FRED) public data API (series \strong{GDPC1}), expressed in billions
of chained 2017 US dollars, seasonally adjusted annual rate.
}
\details{
The dataset covers 1947 Q1 through the latest vintage available at download
time (approximately 316 rows as of 2025). Rows with \code{NA} in \code{gdp_real} are
excluded.

The log-level column (\code{gdp_log}) is particularly useful for trend-cycle
decomposition because log-differences approximate quarter-on-quarter
percentage growth rates:
\deqn{\Delta \log(\text{GDP}_t) \approx g_t}
}
\examples{
data("us_gdp_vintage", package = "MacroFilters")
head(us_gdp_vintage)
plot(us_gdp_vintage$date, us_gdp_vintage$gdp_log,
     type = "l", xlab = "Date", ylab = "Log Real GDP",
     main = "US Real GDP (log level)")
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/05_algo_mbh.R
\name{mbh_filter}
\alias{mbh_filter}
\title{MacroBoost Hybrid (MBH) Filter}
\usage{
mbh_filter(
  x,
  knots = NULL,
  mstop = 500L,
  d = NULL,
  nu = 0.1,
  df = 4L,
  select_mstop = FALSE,
  boundary.knots = NULL
)
}
\arguments{
\item{x}{Numeric vector, \code{ts}, \code{xts}, or \code{zoo} object.}

\item{knots}{Integer.
Number of interior knots for the P-Spline.
If \code{NULL} (default), it is calculated as \code{max(20, floor(n / 2))}.
High knot density is required for the trend to be flexible enough.}

\item{mstop}{Integer.
Maximum number of boosting iterations (default 500).
If \code{select_mstop = TRUE} this is the upper bound; the actual stopping
point is chosen by AICc.}

\item{d}{Numeric or \code{NULL}.
The delta parameter for Huber loss. If \code{NULL} (default), it is
auto-calibrated as the Median Absolute Deviation (MAD) of the first
differences of the series (\code{mad(diff(y))}).

\strong{Scale-mismatch warning:} When \code{x} is a log-level series,
\code{diff(y)} returns inter-period growth rates (typical scale 0.001–0.02),
whereas the output gap (the residual the filter must explain) has a much
larger scale (typical scale 0.01–0.05). Using \code{mad(diff(y))} as \code{d}
therefore sets the Huber threshold far too low: the filter treats normal
business-cycle oscillations as outliers, truncates their gradients, and
blocks learning. The trend becomes over-smooth and the cycle absorbs
too much long-run variance.

\strong{Recommendation:} For a quick preliminary calibration use
\code{d = mad(hp_filter(x)$cycle)}, which sets the threshold on the residual
scale. Supply an explicit numeric value to override the automatic
fallback entirely.}

\item{nu}{Numeric.
The learning rate (shrinkage) for boosting (default 0.1).}

\item{df}{Integer.
Effective degrees of freedom per boosting step for the P-Spline base
learner (default 4). This enforces the \emph{weak-learner} constraint of
Bühlmann & Hothorn (2007): each boosting step contributes only a small,
smooth update so that the trend is built up gradually over many
iterations rather than fitted in one pass.

\strong{End-point instability warning:} Higher \code{df} values cause the B-spline
basis matrix to shift drastically when the sample size changes by even
one observation (the "rubber-band effect"). The last few data points pull
the estimated trend non-smoothly, producing unreliable end-of-sample
estimates. Keep \code{df = 4} (the default) unless you have a specific reason
to deviate.}

\item{select_mstop}{Logical.
If \code{TRUE}, the optimal number of boosting iterations is selected
automatically via AICc (corrected AIC), following Bühlmann & Hothorn
(2007). The \code{mstop} argument acts as the search upper bound. Default
\code{FALSE}.

\strong{AICc underfitting warning:} In the combination of Huber
quasi-likelihood + P-splines, AICc penalises model complexity
hyper-aggressively. In practice the algorithm stops at iteration ~5–15
instead of the intended ~500. The resulting trend is nearly a straight
line; all long-run variance is pushed into the cycle component, defeating
the purpose of the filter. Treat \code{select_mstop = TRUE} as an
experimental option and validate visually before relying on it.}

\item{boundary.knots}{A numeric vector of length 2 specifying the global
domain for the B-spline basis (e.g., \code{c(1, T_max)}). If \code{NULL} (default),
the range of \code{time_idx} is used. For real-time stability, fix this to the
full-sample domain so the basis does not shift as the sample grows.}
}
\value{
A \code{macrofilter} object with \code{trend}, \code{cycle}, \code{data}, and \code{meta}
components.  The \code{meta} list contains \code{method = "MBH"}, \code{knots}, \code{d},
\code{mstop}, \code{nu}, \code{df}, \code{select_mstop}, and \code{compute_time}.
}
\description{
Decomposes a time series into trend and cycle using a robust boosting
algorithm. Unlike the HP filter, MBH uses the Huber loss function to
automatically downweight outliers (like the COVID-19 shock), preventing
them from distorting the trend.
}
\details{
The model estimated is an additive model:
\deqn{y_t = \text{Linear}(t) + \text{Smooth}(t) + \epsilon_t}

It is fitted using \code{\link[mboost:gamboost]{mboost::mboost()}} with:
\itemize{
\item \strong{Base Learners:} A linear time trend (\code{\link[mboost:baselearners]{mboost::bols()}}) to capture
the global path, plus a B-spline (\code{\link[mboost:baselearners]{mboost::bbs()}}) to capture local
curvature.
\item \strong{Loss Function:} Huber loss (\code{\link[mboost:Family]{mboost::Huber()}}) with parameter
\code{d}.  This is the key to robustness.
}

The default parameters (\code{knots = n/2}, \code{mstop = 500}) are calibrated to
mimic the flexibility of a standard HP filter while retaining the robustness
of the Huber loss.
}
\section{Calibration Guidance}{

Three failure modes were discovered through empirical stress-testing.
The defaults guard against all three:

\describe{
\item{1. Huber delta scale mismatch (\code{d})}{
The automatic fallback \code{mad(diff(y))} operates on the scale of
growth rates, not the output gap. For log-level input this sets \code{d}
one to two orders of magnitude too small, causing ordinary
business-cycle swings to be treated as outliers. If the estimated
cycle looks implausibly large or the trend is nearly linear, override
with \code{d = mad(hp_filter(x)$cycle)} as a starting point.
}
\item{2. AICc underfitting (\code{select_mstop})}{
AICc + Huber quasi-likelihood + P-splines stops boosting at
iteration ~5--15. The trend degenerates to a near-straight line and
the cycle absorbs all long-run variance. Leave \code{select_mstop = FALSE}
(the default) and set \code{mstop} explicitly instead.
}
\item{3. End-point instability (\code{df})}{
Values above 4 shift the B-spline basis matrix non-smoothly as
the sample grows, producing a "rubber-band" distortion in the final
observations. Keep \code{df = 4} (the default) for real-time applications.
}
}
}

\examples{
# Fast example with reduced series and iterations
set.seed(42)
y <- ts(cumsum(rnorm(80)), start = c(2000, 1), frequency = 4)
result <- mbh_filter(y, mstop = 100L)
print(result)

\donttest{
# Full example with default parameters
y2 <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
result2 <- mbh_filter(y2)
print(result2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/04_algo_bhp.R
\name{bhp_filter}
\alias{bhp_filter}
\title{Boosted HP Filter}
\usage{
bhp_filter(
  x,
  lambda = NULL,
  iter_max = 100L,
  stopping = c("bic", "adf", "fixed"),
  sig_level = 0.05,
  freq = NULL
)
}
\arguments{
\item{x}{Numeric vector, \code{ts}, \code{xts}, or \code{zoo} object.}

\item{lambda}{Smoothing parameter.
If \code{NULL} (default), it is auto-detected using the Ravn-Uhlig rule
(\code{6.25 * freq^4}).}

\item{iter_max}{Integer.
Maximum number of boosting iterations (default 100).}

\item{stopping}{Character.
Stopping rule: \code{"bic"} (default), \code{"adf"}, or \code{"fixed"}.}

\item{sig_level}{Numeric.
Significance level for the ADF test when \code{stopping = "adf"}
(default 0.05).}

\item{freq}{Numeric frequency override (1 = annual, 4 = quarterly,
12 = monthly). Used only when \code{lambda} is \code{NULL} and the frequency cannot
be inferred from \code{x}.}
}
\value{
A \code{macrofilter} object with \code{trend}, \code{cycle}, \code{data}, and \code{meta}
components.  The \code{meta} list contains \code{method = "bHP"}, \code{lambda},
\code{iterations}, \code{stopping_rule}, and \code{compute_time}.
}
\description{
Iteratively applies the Hodrick-Prescott filter on residuals to better
capture stochastic trends.  At each iteration the HP smoother is applied to
the current residual and the resulting trend increment is added to the
cumulative trend estimate.  Iteration stops according to one of three rules:
BIC minimisation (default), ADF stationarity test on residuals, or a fixed
number of iterations.
}
\details{
The boosted HP filter starts from the standard HP solution and then
re-applies the same HP smoother to the residual (cycle) component.  The
trend increment from each pass is accumulated, and the procedure stops
when one of the following criteria is met:

\describe{
\item{\code{"bic"}}{Schwarz information criterion computed as
\eqn{n \log(\hat\sigma^2) + \log(n)\,\mathrm{tr}(S^m)},
where \eqn{S^m} is the iterated smoother.  Iteration stops when the BIC
increases relative to the previous best.}
\item{\code{"adf"}}{Augmented Dickey-Fuller test on the residual.  Iteration
stops when the residual is stationary at level \code{sig_level}.}
\item{\code{"fixed"}}{Runs exactly \code{iter_max} iterations.}
}
}
\examples{
# Quarterly GDP-like series
y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
result <- bhp_filter(y)
print(result)
}
\references{
Phillips, P.C.B. and Shi, Z. (2021). Boosting: Why You Can Use the HP
Filter. \emph{International Economic Review}, 62(2), 521--570.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/03_algo_hamilton.R
\name{hamilton_filter}
\alias{hamilton_filter}
\title{Hamilton Filter}
\usage{
hamilton_filter(x, h = NULL, p = 4L)
}
\arguments{
\item{x}{Numeric vector, \code{ts}, \code{xts}, or \code{zoo} object.}

\item{h}{Integer horizon (number of periods ahead). If \code{NULL} (default),
auto-detected from the series frequency using Hamilton's rule:
annual = 2, quarterly = 8, monthly = 24.}

\item{p}{Integer number of lags in the regression (default 4).}
}
\value{
A \code{macrofilter} object with \code{trend}, \code{cycle}, \code{data}, and \code{meta}
components. \code{meta} includes \code{h}, \code{p}, \code{coefficients}, and \code{compute_time}.
}
\description{
Decomposes a time series into trend and cycle components using the
regression-based filter proposed by Hamilton (2018). The trend is the
fitted value from an OLS regression of \eqn{y_{t+h}} on
\eqn{(1, y_t, y_{t-1}, \ldots, y_{t-p+1})}, and the cycle is the residual.
}
\details{
Hamilton (2018) proposes replacing the HP filter with a simple regression:
\deqn{y_{t+h} = \beta_0 + \beta_1 y_t + \beta_2 y_{t-1} + \cdots +
  \beta_p y_{t-p+1} + v_{t+h}}
The fitted values \eqn{\hat{y}_{t+h}} define the trend and the residuals
\eqn{\hat{v}_{t+h}} define the cycle.

The first \eqn{h + p - 1} observations have no computable trend or cycle
and are filled with \code{NA}.

The lag matrix is constructed vectorized via \code{embed()} and the
regression is solved with \code{\link[stats:lmfit]{stats::lm.fit()}} for speed.
}
\examples{
# Quarterly GDP-like series
y <- ts(cumsum(rnorm(200)), start = c(2000, 1), frequency = 4)
result <- hamilton_filter(y)
print(result)
}
\references{
Hamilton, J.D. (2018). Why You Should Never Use the Hodrick-Prescott
Filter. \emph{Review of Economics and Statistics}, 100(5), 831--843.
}
